#!/usr/bin/env python

"""...

"""


from typing import Sequence, Mapping, Any, TypeVar
from toyplot.mark import Mark
import toyplot
from toytree.drawing.src.cartesian import Cart
from toytree.drawing.src.cartesian_api import add_cartesian_method

Color = TypeVar("Color")
__all__ = [
    "text",
]


def _mark_exportable(table, column, exportable=True):
    table.metadata(column)["toyplot:exportable"] = exportable


@add_cartesian_method(Cart)
def text(
    axes: Cart,
    a: Sequence[float],
    b: Sequence[float],
    text: Sequence[str],
    angle: float = 0,
    color: Color = None,
    opacity: float = 1.0,
    title: str = None,
    style: Mapping[str, Any] = None,
    filename: str = None,
    annotation: bool = True,
) -> Mark:
    """Add text to the axes.

    Parameters
    ----------
    a, b: float
      Coordinates of the text anchor.
    text: string
      The text to be displayed.
    title: string, optional
      Human-readable title for the mark.  The SVG / HTML backends render the
      title as a tooltip.
    style: dict, optional
      Collection of CSS styles to apply to the mark.  See
      :class:`toyplot.mark.Text` for a list of useful styles.
    annotation: boolean, optional
      Set to True if this mark should be considered an annotation.

    Returns
    -------
    text: :class:`toyplot.mark.Text`
    """
    table = toyplot.data.Table()
    table["x"] = toyplot.require.scalar_vector(a)
    table["y"] = toyplot.require.scalar_vector(b, table.shape[0])
    table["text"] = toyplot.broadcast.pyobject(text, table.shape[0])
    _mark_exportable(table, "x")
    _mark_exportable(table, "y")
    _mark_exportable(table, "text")

    table["angle"] = toyplot.broadcast.scalar(angle, table.shape[0])
    table["opacity"] = toyplot.broadcast.scalar(opacity, table.shape[0])
    table["title"] = toyplot.broadcast.pyobject(title, table.shape[0])
    style = toyplot.style.require(style, allowed=toyplot.style.allowed.text)

    default_color = [next(axes._text_colors)]

    color = toyplot.color.broadcast(
        colors=color,
        shape=(table.shape[0], 1),
        default=default_color,
    )
    table["fill"] = color[:, 0]

    mark = toyplot.mark.Text(
        coordinate_axes=["x", "y"],
        table=table,
        coordinates=["x", "y"],
        text=["text"],
        angle=["angle"],
        fill=["fill"],
        opacity=["opacity"],
        title=["title"],
        style=style,
        annotation=annotation,
        filename=filename,
    )
    return axes.add_mark(mark)


if __name__ == "__main__":

    import toytree

    canvas = toyplot.Canvas()
    axes = Cart(scenegraph=canvas._scenegraph)
    canvas._scenegraph.add_edge(canvas, "render", axes)
    m0 = axes.text(0, 0, "hello world")
    m1 = axes.text(1, 1, "hello world")

    toytree.utils.show(canvas)



#!/usr/bin/env python

"""Neighbor-joining distance based tree inference method.

Algorithm is available in both a performant version (...) as well 
as a didactic version (...).

TODO
----
Implement BioNJ algorithm.
 Gascuel, O (1997) "BIONJ: an improved version of the NJ 
 algorithm based on a simple model of sequence data."

References
----------
...
"""

from typing import TypeVar
from numpy.typing import ArrayLike
import numpy as np
from numba import njit


ToyTree = TypeVar("ToyTree")


def infer_neigbor_joining_tree(matrix: ArrayLike) -> ToyTree:
    """Return a ToyTree inferred by NeighborJoining on a distance matrix.

    Parameters
    ----------
    matrix: ArrayLike
        A symmetric matrix (e.g., ndarray or dataframe) with float or
        int values representing distances between samples. If dataframe
        the returned tree will use the index labels as tip names, else
        the tip labels will be the numeric row indices.

    Examples
    --------
    ...

    References
    ----------
    ...
    """
    # store current nodes indices: [0, 1, 2, 3, ...]. This will 
    # initially contain only tips, but later internal node idxs also.
    idxs = np.arange(matrix.shape[0])
    
    # create a copy of original matrix as float type
    arr = np.array(matrix).astype(np.float64)
    
    # iteratively reduce matrix dims by joining nodes until shape=3
    nodes = []
    while mat.shape[0] > 3:
        
        # cluster nodes with shortest distance, returns node indices,
        # distances, and new matrix with its node indices.
        fidx, gidx, dist_fu, dist_gu, mat, idxs = nj_reduce_matrix_jit(mat, idxs)

        # store info about Nodes that were clustered.
        nodes.append([fidx, gidx, dist_fu, dist_gu, max(idxs)])
    
    # final dist: get dist from internal to newnode
    v0 = (1 / 2.) * mat[0, -1]   
    v1 = (1 / (2 * (3 - 2))) * (mat[-1, :].sum() - mat[0, :].sum())
    dist_vw = v0 + v1
    dist_dw = mat[0, -1] - dist_vw
    dist_ew = mat[1, -1] - dist_vw
    nodes.append([idxs[0], idxs[1], dist_dw, dist_ew, dist_vw])
    
    # build tree from nodes info
    tree = build_tree(nodes)

    # relabel tips with dataframe indices if input was a dataframe
    if isinstance(matrix, pd.DataFrame):
        tree = tree.set_node_data(
            feature="name", 
            mapping=dict(enumerate(matrix.index)),
        )
    return tree


@njit
def get_new_matrix_jit(
    mat: ArrayLike, mask: ArrayLike, f: float, g: float) -> ArrayLike:
    """Return matrix w/ distance of all nodes to the new node. 

    The returned matrix has dimension - 1 relative to original.
    """
    # get new matrix with dim-1 to fill
    ntips = mat.shape[0] - 1
    newmat = np.zeros((ntips, ntips))
    
    # fill newmat with dist between remaining tips
    newmat[:ntips - 1, :ntips - 1] = mat[mask, :][:, mask]  
    
    # fill final row/col with dists to newnode
    val = (1 / 2.) * (mat[f][mask] + mat[g][mask] - mat[f, g])
    newmat[-1, :-1] = val
    newmat[:-1, -1] = val 
    return newmat
        
@njit
def get_q_matrix_jit(mat: ArrayLike) -> ArrayLike:
    """Return q-matrix w/ distance of each tip to all other nodes.

    $$
    Q(i,j) = (n-2)d(i,j) - Sum(d(i,k) - Sum(d(j,k)
    $$
    """
    # create an empty matrix of same size
    qmat = np.zeros(mat.shape)
    nrows = mat.shape[0]

    # iterate over all cells in the lower-left triangle
    for i in range(nrows):
        for j in range(i + 1, nrows):
            if i != j:
                # fill values 
                val0 = (nrows - 2) * mat[i, j]
                val1 = mat[i, :].sum() 
                val2 = mat[j, :].sum()
                val = val0 - val1 - val2
                qmat[i, j] = val
                qmat[j, i] = val
    return qmat

@njit
def get_new_dist_jit(mat: ArrayLike, f: float, g: float) -> float:
    """Return distance from tips (f, g) to new node."""
    n = mat.shape[0]
    v0 = (1 / 2.) * mat[f, g]
    v1 = 1 / (2. * (n - 2))
    v2 = mat[f, :].sum()
    v3 = mat[g, :].sum()
    val = v0 + v1 * (v2 - v3)
    return val

@njit
def nj_reduce_matrix_jit(mat: ArrayLike, idxs: ArrayLike) -> ArrayLike:
    """Returns one agglomeration of the shortest distance in matrix.
    """
    # the q-matrix compares all-by-all
    qmat = get_q_matrix_jit(mat)

    # join the nodes with lowest q-score to create new node u
    min_idxs = np.where(qmat == qmat.min())
    f = min_idxs[0][0]
    g = min_idxs[1][0]

    # get distance of new node from children
    dist_fu = get_new_dist_jit(mat, f, g)
    dist_gu = mat[f, g] - dist_fu

    # get mask to hide rows and columns of agglomerated indices
    mask0 = np.arange(mat.shape[0]) != f
    mask1 = np.arange(mat.shape[0]) != g
    mask = mask0 & mask1

    # create new matrix removing (f, g) and adding (u)
    mat = get_new_matrix_jit(mat, mask, f, g)

    # get index labels
    fidx = idxs[f]
    gidx = idxs[g]

    # reset labels
    nidxs = np.zeros(mat.shape[0], dtype=np.int64)
    nidxs[:-1] = idxs[mask]
    nidxs[-1] = max(idxs) + 1
    idxs = nidxs

    # return f, g, u, fdist, gdist, nidx
    return fidx, gidx, dist_fu, dist_gu, mat, idxs

def build_tree_from_xxx(nodes):
    """Return a ToyTree constructed from ...

    Build a newick string and parse into a tree. Nodes are nested
    in order in the nodes list. Each element contains:
    [node_index0, node_index1, dist0, dist1, parent_index]
    except the final element in which the last element is the 
    internal edge length.
    """
    # store internal node indices to newick strings
    relate = {}
    
    # iterate over internal nodes in order
    for n in nodes[:-1]:
        
        # expand internal nodes to newick strings
        if n[0] in relate:
            n[0] = relate[n[0]]
        if n[1] in relate:
            n[1] = relate[n[1]]
            
        # make newick of the new node pair
        newick = "({}:{},{}:{})".format(n[0], n[2], n[1], n[3])
        
        # store internal node 
        relate[n[4]] = newick
    
    # create final unrooted polytomy
    n0, n1, dist0, dist1, iedge = nodes[-1]    
    newick = "({}:{}, {}:{}, {}:{});".format(newick, iedge, n0, dist0, n1, dist1)
    return toytree.tree(newick)



if __name__ == "__main__":

    import pandas as pd
    import toyplot.browser

    MAT = pd.DataFrame(
        columns=list("abcde"),
        index=list("abcde"),
        data=[
            [0, 5, 9, 9, 8],
            [5, 0, 10, 10, 9],
            [9, 10, 0, 8, 7],
            [9, 10, 8, 0, 3],
            [8, 9, 7, 3, 0],
        ],
    )

    # piecewise example
    IDXS = np.arange(MAT.shape[0])
    ARR = np.array(MAT).astype(np.float64)

    print(IDXS)
    print(ARR)

    fidx, gidx, dist_fu, dist_gu, mat, idxs = nj_reduce_matrix_jit(ARR, IDXS)
    print(fidx, gidx, dist_fu, dist_gu, mat, idxs)

    # full example
    # TREE = infer_neigbor_joining_tree(MAT)
    # TREE._draw_browser()




#!/usr/bin/env python

"""Test get_node_data 

- normal use
- missing data
- complex dtypes
- complex mixed dtypes
"""

import unittest
import numpy as np
import toytree


class TestNodeData(unittest.TestCase):

    def setUp(self):
        self.tree = toytree.rtree.unittree(6)

    def test_int_data(self):
        self.tree.set_node_data("x", range(self.tree.nnodes), inplace=True)
        self.assertListEqual(
            list(self.tree.get_node_data('x').values),
            list(range(self.tree.nnodes))
        )

    def test_int_data_w_missing(self):
        """This will convert the ints to floats in get_node_data."""
        self.tree.set_node_data("x", {5: 5, 6: 6}, inplace=True)
        self.assertListEqual(
            list(self.tree.get_node_data('x').values),
            [np.nan, np.nan, np.nan, np.nan, np.nan, 5, 6, np.nan, np.nan, np.nan, np.nan]
        )


    def test_get_node_data(self):
        pass


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""Tree Styles checked and expanded at .draw() or repr().

Examples
--------
>>> toytree.tree(...)                   # DefaultTreeStyle() init
>>> tree.draw(**kwargs)                 # dict or user args
>>> style = DefaultTreeStyle(**kwargs)  # DefaultTreeStyle(**dict, **tree.style)
"""

from typing import List, Tuple, Optional, Union, Iterable, Dict, TypeVar
import json
from enum import Enum
from dataclasses import dataclass, asdict, field
import toyplot
from loguru import logger
import numpy as np
import pandas as pd
# from toytree.color import ToyColor, Color, color_parser, color_cycler
from toytree.color import ToyColor, color_cycler, ColorType
from toytree.utils.transform import normalize_values
# from toytree.utils import ToytreeError

ToyTree = TypeVar("ToyTree")

class EdgeType(str, Enum):
    CLADOGRAM = 'c'
    PHYLOGRAM = 'p'
    BEZIER = 'b'

class LayoutType(str, Enum):
    RIGHT = 'right'
    LEFT = 'left'
    DOWN = 'down'
    UP = 'up'
    CIRCLE = 'circular'
    #unrooted = 'unrooted'

class SubStyle:
    """A subclass of Style for CSS on markers.

    When converted to dict the keys of substyles will be changed
    from e.g., _toyplot_anchor_shift to -toyplot-anchor-shift.
    """

@dataclass
class NodeStyle(SubStyle):
    fill: ColorType = 'rgba(40.0%,76.1%,64.7%,1.000)'
    fill_opacity: Optional[float] = None
    stroke: Optional[ColorType] = "#262626"
    stroke_width: float = 1.0

@dataclass
class NodeLabelStyle(SubStyle):
    fill: ColorType = 'rgba(16.1%,15.3%,14.1%,1.000)'
    font_size: Union[int, str] = 9
    font_weight: int = 300
    font_family: str = "Helvetica"
    _toyplot_anchor_shift: Union[str, int] = 0
    baseline_shift: Union[str,int] = 0
    text_anchor: str = "middle"

@dataclass
class EdgeStyle(SubStyle):
    stroke: ColorType = 'rgba(16.1%,15.3%,14.1%,1.000)'
    stroke_width: float = 2.
    stroke_opacity: Optional[float] = None
    stroke_linecap: str = "round"
    stroke_dasharray: Optional[str] = None

@dataclass
class EdgeAlignStyle(SubStyle):
    stroke: ColorType = 'rgba(66.3%,66.3%,66.3%,1.000)'
    stroke_width: int = 2
    stroke_opacity: Optional[float] = 0.75
    stroke_linecap: str = "round"
    stroke_dasharray: str = "2,4"

@dataclass
class TipLabelsStyle(SubStyle):
    fill: ColorType = 'rgba(16.1%,15.3%,14.1%,1.000)'
    fill_opacity: Optional[float] = None
    font_size: Union[str, float] = 11
    font_weight: int = 300
    font_family: str = "Helvetica"
    _toyplot_anchor_shift: Union[str, float] = 15
    baseline_shift: Union[str,int] = 0
    text_anchor: str = "start"

######################################################################
#
#  Base TreeStyle Class with validation functions
#
######################################################################

@dataclass
class TreeStyle:
    """TreeStyle is a dict-like object for setting/getting style args.

    TreeStyle validates input style arguments; expands shortcut
    references to styles before calls to draw; and can serialize
    style args for displaying when .style is called interactively.

    Example
    -------
    >>> tree = toytree.rtree.unittree(10)
    >>> tree.style.edge_widths = 4
    >>> tree.draw();

    Note
    ----
    The function `._validate()` takes args for whether to serialize

    tree.style._validate(tree)
    """
    # inherited from tree at validation steps
    tree_style: Optional[str] = None
    height: Optional[int] = None
    width: Optional[int] = None
    layout: LayoutType = 'r'

    edge_type: EdgeType = 'p'
    edge_colors: Union[ColorType, Iterable[ColorType], None] = None  # override by fill
    edge_widths: Union[int, Iterable[int], None] = None  # override by stroke-width
    edge_style: EdgeStyle = field(default_factory=EdgeStyle)
    edge_align_style: EdgeAlignStyle = field(default_factory=EdgeAlignStyle)

    node_mask: Union[bool, Iterable[bool], None] = None  # None = auto
    node_colors: Union[ColorType, Iterable[ColorType], None] = None  # override by fill
    node_sizes: Union[int, Iterable[int]] = 0
    node_markers: Union[str, Iterable[str]] = "o"
    node_hover: Union[bool, str, Iterable[str]] = None  # None = auto
    node_style: NodeStyle = field(default_factory=NodeStyle)

    node_labels: Union[bool, str, Iterable[str]] = False
    node_labels_style: NodeLabelStyle = field(default_factory=NodeLabelStyle)

    tip_labels: Union[bool, Iterable[str]] = True
    tip_labels_colors: Union[str, Iterable[str], None] = None
    tip_labels_align: bool = False
    tip_labels_style: TipLabelsStyle = field(default_factory=TipLabelsStyle)
    tip_labels_angles: Union[float, Iterable[float], None] = None

    use_edge_lengths: bool = True
    scale_bar: Union[bool,float] = False
    padding: int = 15
    xbaseline: int = 0
    ybaseline: int = 0
    admixture_edges: Union[List[Tuple[int, int]], None] = None
    shrink: int = 0

    def validate(self, tree: ToyTree):
        """Expand style args and convert types.

        Expand iterable style entires to None or ndarray, and
        convert color args to CSS strings. This function is called
        before passing style args to ToyTreeMark.

        The validate x_style funcs should come last in each group
        because the fill,stroke styles can change.

        TODO: make faster,this is always called when?
        """
        self.tree = tree  # ignore
        self._validate_node_colors()
        self._validate_node_mask()
        self._validate_node_sizes()
        self._validate_node_markers()
        self._validate_node_hover()
        self._validate_node_style()

        self._validate_node_labels()
        self._validate_node_labels_style()

        self._validate_tip_labels()
        self._validate_tip_labels_colors()
        self._validate_tip_labels_angles()
        self._validate_tip_labels_style()

        self._validate_edge_colors()
        self._validate_edge_widths()
        self._validate_edge_style()
        self._validate_edge_align_style()

        self._validate_admixture_edges()

    def _serialize_colors(self):
        """Convert all ndarrays to lists for serialization of JSON."""
        self._validate_node_colors()
        self._validate_tip_labels_colors()
        self._validate_edge_colors()
        # ... substyledicts

    def _serialize_css_styles(self):
        """Convert keys with lower_case ndarrays to lists for serialization of JSON..."""
        self._validate_node_colors()
        self._validate_tip_labels_colors()
        # ... substyledicts

    def _serialize_arrays(self):
        """
        Convert any numpy or pandas inputs to lists for easier
        validation casting. This is called for serialization and
        during validation.
        """
        arrayed = [
            "node_mask",
            "node_sizes",
            "node_labels",
            "node_colors",
            "tip_labels",
        ]
        for key in arrayed:
            values = getattr(self, key)
            if isinstance(values, pd.DataFrame):
                logger.warning("DataFrame entry is ambiguous, select a Series")
                setattr(self, key, values.iloc[:, 0].tolist())
            elif isinstance(values, pd.Series):
                setattr(self, key, values.tolist())
            # TODO, maybe this is unnecessary now with color-parser
            elif isinstance(values, np.ndarray):
                # array can be a collection of things, or a color array
                if hasattr(values, 'r'):
                    setattr(self, key, ToyColor(values).css)
                else:
                    setattr(self, key, values.tolist())

    def _validate_node_colors(self) -> None:
        """Set .node_colors and node_style.fill (competing args).

        If node_style.fill then node_colors will be None, else .fill
        be None and node_colors will be expanded to an ndarray[str].
        """
        if self.node_colors is None:
            return
        # convert to a ToyColor or list of ToyColors
        colors = color_parser(self.node_colors)

        if isinstance(colors, ToyColor):
            css_color = colors.css
            self.node_colors = None
            self.node_style.fill = css_color
        else:
            colors = [i.css for i in colors]
            self.node_colors = toyplot.broadcast.pyobject(
                colors, self.tree.nnodes).astype(str)
            self.node_style.fill = None

    def _validate_node_mask(self) -> None:
        """Sets node_mask to ndarray[bool].
            None: special to hide tips only
            True: show all
            False: hide all
            Iterable: custom
        """
        if self.node_mask is True:
            self.node_mask = np.repeat(self.node_mask, self.tree.nnodes)
        if self.node_mask is False:
            self.node_mask = np.repeat(self.node_mask, self.tree.nnodes)
        if self.node_mask is None:
            self.node_mask = np.zeros(self.tree.nnodes, dtype=bool)
            self.node_mask[:self.tree.ntips] = True
        self.node_mask = toyplot.broadcast.pyobject(
            self.node_mask, self.tree.nnodes).astype(bool)

    def _validate_node_sizes(self) -> None:
        """Sets node_sizes to ndarray[float]."""
        self.node_sizes = toyplot.broadcast.scalar(
            self.node_sizes, self.tree.nnodes)

    def _validate_node_markers(self):
        """Sets node_markers to ndarray[str]."""
        self.node_markers = toyplot.broadcast.pyobject(
            self.node_markers, self.tree.nnodes)

    def _validate_node_labels(self):
        """
        Sets node_labels to ndarray[str] or None and apply simple
        floating point string formatting on node_labels.
        """
        # get node_labels as either None or mixed type
        if self.node_labels is False:
            self.node_labels = None
        elif self.node_labels is None:
            self.node_labels = None
        elif self.node_labels is True:
            self.node_labels = range(self.tree.nnodes)
        elif isinstance(self.node_labels, pd.Series):
            # assume user can float format themselves...
            pass
        elif isinstance(self.node_labels, str):
            # auto-float format
            self.node_labels = self.tree.get_node_data(self.node_labels)

            # check if it can be cast from float to int
            if self.node_labels.dtype == float:

                # wrap in try to allow for customs with NaN
                try:
                    # trim off anything after .6, and string format to match
                    self.node_labels = self.node_labels.round(6)

                    # check if it can be cast to int
                    if all(i % 1 == 0 for i in self.node_labels):
                        self.node_labels = self.node_labels.astype(int)
                except Exception:
                    pass

        # double check size and cast to str
        if self.node_labels is not None:
            self.node_labels = toyplot.broadcast.pyobject(
                self.node_labels, self.tree.nnodes).astype(str)

    def _validate_node_labels_style(self):
        """
        Ensure tip labels are in px sizes and check fill color
        """
        self.node_labels_style.font_size = "{:.1f}px".format(
            toyplot.units.convert(
                value=self.node_labels_style.font_size,
                target="px", default="px")
        )
        if self.node_labels_style.fill is not None:
            self.node_labels_style.fill = ToyColor(self.node_labels_style.fill)

    def _validate_node_hover(self):
        """Sets node_hover to ndarray[str] or None.

        No comparisons use 'is in' to support flexible input types
        including pd.Series.
        """
        if self.node_hover is None:
            self.node_hover = None
        elif self.node_hover is False:
            self.node_hover = None
        elif self.node_hover is True:
            ordered_features = ["idx", "dist", "support", "height"]
            lfeatures = list(set(self.tree.features) - set(ordered_features))
            ordered_features += lfeatures
            self.node_hover = [" "] * self.tree.nnodes
            for idx in range(self.tree.nnodes):
                feats = []
                node = self.tree[idx]
                for feature in ordered_features:
                    val = getattr(node, feature)
                    if isinstance(val, float):
                        fstring = np.format_float_positional(round(val, 8), trim='0')
                        feats.append(f"{feature}: {fstring}")
                    else:
                        feats.append(f"{feature}: {val}")
                self.node_hover[idx] = "\n".join(feats)
        elif isinstance(self.node_hover, pd.Series):
            self.node_hover = self.node_hover.astype(str)
        elif isinstance(self.node_hover, str):
            self.node_hover = self.tree.get_node_data(self.node_hover).astype(str)
        else:
            self.node_hover = toyplot.broadcast.pyobject(self.node_hover, self.tree.nnodes).astype(str)

    def _validate_node_style(self):
        """Sets node_style.fill and .stroke to str"""
        if self.node_style.stroke is None or self.node_style.stroke == "none":
            self.node_style.stroke = "transparent"
        if self.node_style.stroke is not None:
            self.node_style.stoke = ToyColor(self.node_style.stroke).css
        if self.node_style.fill == "none":
            self.node_style.fill = None
        if self.node_style.fill is not None:
            self.node_style.fill = ToyColor(self.node_style.fill).css

    def _validate_tip_labels_colors(self):
        """Sets tip_labels_colors to ndarray[str] or None, and sets
        tip_labels.style.fill to None or single color.
        """
        if self.tip_labels_colors is None:
            return
        # convert to a ToyColor or list of ToyColors
        colors = color_parser(self.tip_labels_colors)
        if isinstance(colors, ToyColor):
            css_color = colors.css
            self.tip_labels_colors = None
            self.tip_labels_style.fill = css_color
        else:
            colors = [i.css for i in colors]
            self.tip_labels_colors = toyplot.broadcast.pyobject(
                colors, self.tree.ntips).astype(str)
            self.tip_labels_style.fill = None

    def _validate_tip_labels(self):
        """Expand tip labels to a list of strings of len=ntips."""
        if self.tip_labels is True:
            self.tip_labels = np.array(self.tree.get_tip_labels())
        elif self.tip_labels is False:
            self.tip_labels = None
        else:
            if isinstance(self.tip_labels, str):
                if self.tip_labels in self.tree.features:
                    self.tip_labels = self.tree.get_tip_data(self.tip_labels)
            self.tip_labels = toyplot.broadcast.pyobject(
                self.tip_labels, self.tree.ntips).astype(str)

    def _validate_tip_labels_angles(self):
        """Sets tip_labels_angles to ndarray[float]. None is auto layout."""

        # udlr sets simple. c and unr auto sets during layout, unless
        # user enters an argument for tip_lables_angles to draw.
        if self.tip_labels_angles is None:
            if self.layout in ['u', 'd']:
                angles = -90
            else:
                angles = 0
        # use auto value filled during layout, or user arg.
        else:
            angles = self.tip_labels_angles
        self.tip_labels_angles = (
            toyplot.broadcast.scalar(angles, self.tree.ntips))

    def _validate_tip_labels_style(self):
        """Convert user font-size arg to px units."""
        size = toyplot.units.convert(
            value=self.tip_labels_style.font_size,
            target="px", default="px"
        )
        self.tip_labels_style.font_size = f"{size:.1f}px"
        # self.tip_labels_style._toyplot_anchor_shift = "{:.2f}px".format(
        #     toyplot.units.convert(
        #         self.tip_labels_style._toyplot_anchor_shift, "px", "px")
        # )

    def _validate_edge_colors(self):
        """Resolve competing args for edge colors.

        If edge_style.stroke is not None then a single value is used
        and edge_colors is set to None, else edge_colors will be
        expanded into an ndarray[str].

        TODO: Allow values to be nnodes or nnodes -1, since root edge is not shown?
        """
        if self.edge_colors is None:
            return
        # convert to a ToyColor or list of ToyColors
        colors = color_parser(self.edge_colors)
        if isinstance(colors, ToyColor):
            css_color = colors.css
            self.edge_colors = None
            self.edge_style.stroke = css_color
        else:
            colors = [i.css for i in colors]
            self.edge_colors = toyplot.broadcast.pyobject(
                colors, self.tree.nnodes).astype(str)
            self.edge_style.stroke = None

    def _validate_edge_style(self):
        """
        If edge_colors are variable then edge_style.stroke is set to
        None, else edge_style.stroke is one color. Opacity is set to
        None by default, if a value is set then it applies on top of
        existing colors.
        """
        if isinstance(self.edge_style.stroke, str):
            if self.edge_style.stroke == "none":
                self.edge_style.stroke = None
        if self.edge_style.stroke is not None:
            self.edge_style.stoke = ToyColor(self.edge_style.stroke).css
        # self.edge_style.stroke_opacity

    def _validate_edge_widths(self):
        """
        Sets edge_widths to ndarray[float] or None, in which case the
        value is taken from edge_style.stroke-width.
        """
        if self.edge_widths is not None:
            if isinstance(self.edge_widths, str):
                if self.edge_widths in self.tree.features:
                    self.edge_widths = normalize_values(
                        self.tree.get_node_data(self.edge_widths, missing=2)
                    )
                else:
                    self.edge_widths = 2
        self.edge_widths = toyplot.broadcast.scalar(
            self.edge_widths, self.tree.nnodes)

    def _validate_edge_align_style(self):
        """

        """
        # if self.edge_align_style.stroke == "none":
        #     self.edge_style.stroke = None
        # if self.edge_style.stroke is not None:
        #     self.edge_style.stoke = ToyColor(self.edge_style.stroke).css
        # # self.edge_style.stroke_opacity

    def _validate_admixture_edges(self):
        """Expand admixture args to a list of tuples.

        The source and dest Nodes can be selected using either their
        Node int idx label, or Node name str, or multiple Node name
        strs, the latter of which will select the MRCA Node.

        The proper format should be:

        >>> admixture_edges = [
        >>>     (src_idx, dest_idx, (src_time, dest_time), dict, str)
        >>> ]
        """
        # bail if empty
        if self.admixture_edges is None:
            return

        # if tuple then expand into a list
        if isinstance(self.admixture_edges, tuple):
            self.admixture_edges = [self.admixture_edges]

        # get color generator and skip the first
        icolors = color_cycler()
        admix_tuples = []
        for atup in self.admixture_edges:

            # required: src node idx from Union[int, str, Iterable[str]]
            if isinstance(atup[0], (int, str)):
                src = self.tree.get_mrca_node(atup[0]).idx
            else:
                src = self.tree.get_mrca_node(*atup[0]).idx

            # required: dest node idx from Union[int, str, Iterable[str]]
            if isinstance(atup[1], (int, str)):
                dest = self.tree.get_mrca_node(atup[1]).idx
            else:
                dest = self.tree.get_mrca_node(*atup[1]).idx

            # optional: proportion on edges
            if len(atup) > 2:
                if isinstance(atup[2], (int, float)):
                    prop = float(atup[2])
                else:
                    prop = (float(atup[2][0]), float(atup[2][1]))
            else:
                prop = 0.5

            # optional: style dictionary
            if len(atup) > 3:
                style = dict(atup[3])
            else:
                style = {}

            # optional label on midpoint
            if len(atup) > 4:
                label = str(atup[4])
            else:
                label = None

            # color setting.
            stroke = (
                ToyColor(next(icolors)) if "stroke" not in style
                else ToyColor(style['stroke'])
            )
            style['stroke'] = stroke.color.css
            style['stroke-opacity'] = style.get('stroke-opacity', 0.7)

            # check styledict colors, etc
            admix_tuples.append((src, dest, prop, style, label))
        self.admixture_edges = admix_tuples

    def dict(
        self,
        serialize_colors: bool=False,
        serialize_arrays: bool=False,
        validate_css: bool=False,
        ):
        """Return a (serializable) dictionary of the TreeStyle.

        Parameters
        ----------
        serialize_colors: bool
            Converts colors to CSS strings.
        serialize_arrays: bool
            Converts ndarrays to lists.
        validate_css: bool
            Converts substyle keys to valide CSS (e.g., '_' -> '-'')
        """
        # first convert any ndarray colors to CSS str or Collection[str]
        # whether entered as a single value or a Collectin of values.
        if serialize_colors:
            self._serialize_colors()

        # first converts pandas/numpy to List[Any] or str. This is

        # not called by validate()
        if serialize_arrays:
            self._serialize_arrays()
        # next convert
        style_dict = asdict(self)
        if validate_css:
            self._to_css_styles(style_dict)
        return style_dict

    def json(self):
        """Returns a JSON serialized dict of the TreeStyle"""
        mydict = self.dict(True, True, True)
        return json.dumps(mydict, indent=4)

    def __repr__(self):
        """Recursive dict to JSON representation"""
        return self.json()

    def _to_css_styles(self, style_dict: Dict[str,str]):
        """Convert dict keys to valid CSS styles."""
        style_dict_keys = list(style_dict.keys())
        for key in style_dict_keys:
            sub_style = getattr(self, key)
            if isinstance(sub_style, SubStyle):
                sub_style_dict = style_dict[key]
                sub_style_dict_keys = list(sub_style_dict.keys())
                for skey in sub_style_dict_keys:
                    new_skey = skey.replace("_", "-")
                    svalue = sub_style_dict.pop(skey)
                    sub_style_dict[new_skey] = svalue


######################################################################
#
# BUILTIN TREE STYLES
#
######################################################################


@dataclass(repr=False)
class NormalTreeStyle(TreeStyle):
    tree_style: str = 'n'

@dataclass(repr=False)
class SimpleTreeStyle(TreeStyle):
    tree_style: str = 's'
    node_labels: bool = True
    node_mask: bool = False
    node_sizes: int = 18
    node_style: NodeStyle = field(
        default_factory=lambda: NodeStyle(
            stroke="#262626",
            stroke_width=1.5,
            fill="lightgrey",
        )
    )
    tip_labels: bool = True
    use_edge_lengths: bool = False

@dataclass(repr=False)
class PhyloTreeStyle(TreeStyle):
    tree_style: str = 'p'
    layout: LayoutType = 'd'
    edge_type: EdgeType = 'c'
    edge_widths: Union[int, Iterable[int], str] = "Ne"  # special
    node_labels: bool = True
    node_sizes: int = 15
    node_hover: bool = False
    node_mask: bool = False
    node_style: NodeStyle = field(
        default_factory=lambda: NodeStyle(
            stroke="#262626",
            stroke_width=1.0,
        )
    )
    node_labels_style: NodeLabelStyle = field(
        default_factory=lambda: NodeLabelStyle(
            font_size=9,
        )
    )
    tip_labels: bool = True
    tip_labels_align: bool = False
    scale_bar: bool = True
    use_edge_lengths: bool = True

@dataclass(repr=False)
class UmlautTreeStyle(TreeStyle):
    tree_style: str = "o"
    edge_type: EdgeType = 'c'
    node_sizes: int = 8
    tip_labels: bool = True
    tip_labels_align: bool = True
    edge_style: EdgeStyle = field(
        default_factory=lambda: EdgeStyle(
            stroke='#262626',
            stroke_width=2,
        )
    )
    node_style: NodeStyle = field(
        default_factory=lambda: NodeStyle(
            stroke="white",
            stroke_width=1.5,
            fill='rgba(10.6%,62.0%,46.7%,1.000)',
        )
    )

@dataclass(repr=False)
class CoalTreeStyle(TreeStyle):
    tree_style: str = "c"
    edge_type: EdgeType = 'c'
    layout: LayoutType = 'd'
    scale_bar: bool = True
    node_sizes: int = 7
    node_mask: bool = False
    # tip_labels: str = "idx"
    tip_labels_style: TipLabelsStyle = field(
        default_factory=lambda: TipLabelsStyle(
            _toyplot_anchor_shift = 12,
        )
    )
    edge_style: EdgeStyle = field(
        default_factory=lambda: EdgeStyle(
            stroke='#262626',
            stroke_width=2,
        )
    )
    node_style: NodeStyle = field(
        default_factory=lambda: NodeStyle(
            stroke="#262626",
            stroke_width=1.5,
            fill='rgba(40.0%,76.1%,64.7%,1.000)',
        )
    )

@dataclass(repr=False)
class DarkTreeStyle(TreeStyle):
    tree_style: str = "d"
    tip_labels: bool = True
    tip_labels_style: TipLabelsStyle = field(
        default_factory=lambda: TipLabelsStyle(
            _toyplot_anchor_shift=8,
            fill='rgba(90.6%,54.1%,76.5%,1.000)'
        )
    )
    edge_style: EdgeStyle = field(
        default_factory=lambda: EdgeStyle(
            stroke='rgba(40.0%,76.1%,64.7%,1.000)',
            stroke_width=2,
        )
    )
    node_labels_style: NodeLabelStyle = field(
        default_factory=lambda: NodeLabelStyle(
            fill='rgba(98.8%,55.3%,38.4%,1.000)',
        )
    )
    node_style: NodeStyle = field(
        default_factory=lambda: NodeStyle(
            stroke=None,
            stroke_width=1.5,
            fill='rgba(40.0%,76.1%,64.7%,1.000)', #'white',
        )
    )

# @dataclass(repr=False)
# class UnrootedTreeStyle(TreeStyle):
#     tree_style: str = "u"
#     layout: str = "unrooted"
#     use_edge_lengths: bool = False
#     node_style: NodeStyle = field(
#         default_factory=lambda: NodeStyle(
#             stroke=None,
#             stroke_width=1.5,
#             fill='white',
#         )
#     )
#     node_sizes: int = 6
#     node_labels: str = "idx"


STYLE_DICTS = {
    "n": NormalTreeStyle,
    "s": SimpleTreeStyle,
    "p": PhyloTreeStyle,
    "o": UmlautTreeStyle,
    "c": CoalTreeStyle,
    "d": DarkTreeStyle,
}


def get_base_style_from_name(tree_style: str="n") -> TreeStyle:
    """Return a base TreeStyle indexed by unique str name prefix
    """
    tree_style = tree_style.lower()[0]
    style = STYLE_DICTS[tree_style]
    return (style)()

import time
if __name__ == "__main__":


    start = time.time()
    ts = get_base_style_from_name('s')
    print(ts)
    print(time.time() - start)

    # import toytree
    # import numpy as np

    # TREE = toytree.rtree.unittree(10)
    # tstyle = NormalTreeStyle(
    #     node_labels="hi", #np.zeros(10),
    #     node_hover="yo",
    #     height=400,
    #     width=400,
    #     edge_type="c",
    #     node_sizes=np.zeros(5),
    #     node_colors=list(toyplot.color.Palette()),
    #     edge_colors='red',
    # )
    # tstyle = UmlautTreeStyle(node_colors='red')
    # print(tstyle)



#!/usr/bin/env python

"""
Random Tree generation Class
"""

import numpy as np
import random
import toytree
from .utils import ToytreeError


class RandomTree(object):


    @staticmethod
    def unittree(ntips, treeheight=1.0, random_names=False, seed=None):
        """
        Returns a random tree ultrametric topology w/ N tips and a root 
        height set to 1 or a user-entered treeheight value. Descendant 
        nodes are evenly spaced between the root and time 0.

        Parameters
        -----------
        ntips (int):
            The number of tips in the randomly generated tree

        treeheight(float):
            Scale tree height (all edges) so that root is at this height.

        seed (int):
            Random number generator seed.
        """
        # seed generator
        random.seed(seed)

        # generate tree with N tips.
        tmptree = toytree.tree().treenode  # TreeNode()
        tmptree.populate(ntips)
        self = toytree.tree(newick=tmptree.write())

        # set tip names by labeling sequentially from 0
        self = (
            self
            .ladderize()
            .mod.make_ultrametric()
            .mod.node_scale_root_height(treeheight)
        )

        # set tipnames randomly (doesn't have to match idx)
        nidx = list(range(self.ntips))[::-1]
        if random_names:
            random.shuffle(nidx)
        for tidx, node in enumerate(self.treenode.get_leaves()):
            node.name = "r{}".format(nidx[tidx])

        # set all support values to 100 default
        for node in self.treenode.traverse():
            node.support = 100

        # fill internal node names and idxs
        self.treenode.ladderize()
        self._coords.update()
        return self


    @staticmethod
    def imbtree(ntips, treeheight=1.0, random_names=False):
        """
        Return an imbalanced (comb-like) tree topology.
        """
        node = toytree.TreeNode.TreeNode()
        node.add_child(name="r0")
        node.add_child(name="r1")

        for i in range(2, ntips):
            # empty node
            cherry = toytree.TreeNode.TreeNode()
            # add new child
            cherry.add_child(name="r" + str(i))
            # add old tree
            cherry.add_child(node)
            # update rtree
            node = cherry

        # get toytree from newick            
        tre = toytree.tree(node)
        tre = tre.mod.make_ultrametric(nocopy=True)
        tre = tre.mod.node_scale_root_height(treeheight, nocopy=True)
        tre._coords.update()

        # randomize tip names
        nidx = list(range(tre.ntips))
        if random_names:
            random.shuffle(nidx)
        for idx, node in tre.idx_dict.items():
            if node.is_leaf():
                node.name = "r{}".format(nidx[idx])
        return tre


    @staticmethod
    def baltree(ntips, treeheight=1.0, random_names=False):
        """
        Returns a balanced tree topology.
        """
        # require even number of tips
        if ntips % 2:
            raise ToytreeError("balanced trees must have even number of tips.")

        # make first cherry
        rtree = toytree.tree()
        rtree.treenode.add_child(name="r0")
        rtree.treenode.add_child(name="r1")

        # add tips in a balanced way
        for i in range(2, ntips):

            # get node to split
            node = return_small_clade(rtree.treenode)

            # add two children
            node.add_child(name=node.name)
            node.add_child(name="r" + str(i))

            # rename ancestral node
            node.name = None

        # get toytree from newick            
        tre = toytree.tree(rtree)  # .write(tree_format=9))
        tre = tre.mod.make_ultrametric().mod.node_scale_root_height(treeheight)
        tre._coords.update()

        # rename tips so names are in order
        nidx = list(range(tre.ntips))
        if random_names:
            random.shuffle(nidx)
        for idx, node in tre.idx_dict.items():
            if node.is_leaf():
                node.name = "r{}".format(nidx[idx])
        return tre


    @staticmethod
    def rtree(ntips, treeheight=1.0, random_names=False, seed=None):
        """
        Randomly assigns edge lengths between U(0-1) to edges 
        and then scales total treeheight to 1.0 or entered value.
        """
        # get a unit tree
        self = toytree.rtree.unittree(ntips, treeheight, random_names, seed)

        # randomly assign node dists
        self = self.set_node_values(
            "dist", {i: random.random() for i in range(self.nnodes)}
        )

        # rescale total height to .
        self = self.mod.node_scale_root_height(treeheight, nocopy=True)
        return self


    @staticmethod
    def coaltree(ntips, ne=None, random_names=False, seed=None):
        """
        Returns a coalescent tree with ntips samples and waiting times 
        between coalescent events drawn from the kingman coalescent:
        (4N)/(k*(k-1)), where N is population size and k is sample size.
        Edge lengths on the tree are in generations.

        If no Ne argument is entered then edge lengths are returned in units
        of 2*Ne, i.e., coalescent time units. 
        """
        # seed generator
        random.seed(seed)

        # convert units
        coalunits = False
        if not ne:
            coalunits = True
            ne = 10000

        # build tree: generate N tips as separate Nodes then attach together 
        # at internal nodes drawn randomly from coalescent waiting times.
        tips = [
            toytree.tree().treenode.add_child(name=str(i)) 
            for i in range(ntips)
        ]
        while len(tips) > 1:
            rtree = toytree.tree()
            tip1 = tips.pop(random.choice(range(len(tips))))
            tip2 = tips.pop(random.choice(range(len(tips))))
            kingman = (4. * ne) / float(ntips * (ntips - 1))
            dist = random.expovariate(1. / kingman)
            rtree.treenode.add_child(tip1, dist=tip2.height + dist)
            rtree.treenode.add_child(tip2, dist=tip1.height + dist)
            tips.append(rtree.treenode)

        # build new tree from the newick string
        self = toytree.tree(tips[0].write())    
        self.treenode.ladderize()

        # make tree edges in units of 2N (then N doesn't matter!)
        if coalunits:
            for node in self.treenode.traverse():
                node.dist /= (2. * ne)

        # ensure tips are at zero (they sometime vary just slightly)
        for node in self.treenode.traverse():
            if node.is_leaf():
                node.dist += node.height

        # set tipnames to r{idx}
        nidx = list(range(self.ntips))
        if random_names:
            random.shuffle(nidx)
        for idx, node in self.idx_dict.items():
            if node.is_leaf():
                node.name = "r{}".format(nidx[idx])

        # decompose fills in internal node names and idx
        self._coords.update()
        return self


    @staticmethod
    def bdtree(ntips=10,
                time=4,
                b=1,
                d=0,
                stop="taxa",
                seed=None,
                retain_extinct=False,
                random_names=False,
                verbose=False):
        """
        Generate a classic birth/death tree.

        Parameters
        -----------
        ntips (int):
            Number of tips to generate for 'taxa' stopping criterion.

        time (float):
            Amount of time to simulate for 'time' stopping criterion.

        b (float):
            Birth rate per time unit

        d (float):
            Death rate per time unit (d=0 produces Yule trees)

        stop (str):
            Stopping critereon. Valid values are only 'taxa' or 'time'.

        seed (int):
            Random number generator seed.

        retain_extinct (bool):
            (Unimplemented) Whether to retain internal nodes leading to extinct tips.

        random_names (bool):
            Whether to randomize tip names or name them in order.

        verbose (bool):
            Print some useful information
        """
        if stop not in ["taxa", "time"]:
            raise ToytreeError("stop must be either 'taxa' or 'time'")

        taxa_stop = ntips
        time_stop = time

        # set random seed
        if seed:
            np.random.seed(seed)

        # start from random tree (idxs will be re-assigned at end)
        tre = toytree.tree()
        tre.treenode.idx = 0
        gidx = 1

        # counters for extinctions, total events, and time
        resets = 0
        ext = 0
        evnts = 0
        t = 0

        # continue until stop var
        while 1:

            # get current tips
            tips = tre.treenode.get_leaves()

            # sample time until next event, increment t and evnts
            dt = np.random.exponential(1 / (len(tips) * (b + d)))
            t = t + dt
            evnts += 1

            # sample a [0-1] to choose birth or death and sample a tip node
            r = np.random.sample()
            sp = np.random.choice(tips)

            # event is a birth
            if (r <= b / (b + d)):
                c1 = sp.add_child(name=str(t) + "-1", dist=0)
                c1.add_feature("tdiv", t)
                c1.idx = gidx
                gidx += 1

                c2 = sp.add_child(name=str(t) + "-2", dist=0)
                c2.add_feature("tdiv", t)
                c2.idx = gidx
                gidx += 1

            # else event is extinction
            else:
                # get parent node
                parent = sp.up

                # if no parent then reset to empty tree
                if parent is None:
                    resets += 1
                    tre = toytree.tree()
                    tre.treenode.idx = 0
                    gidx = 1
                    ext = 0
                    evnts = 0
                    t = 0

                # connect parent to sp' children
                else:
                    # if parent is None then reset
                    if sp.up is None:
                        tre = toytree.tree()
                        tre.treenode.idx = 0
                        gidx = 1
                        ext = 0
                        evnts = 0
                        t = 0

                    # if parent is root then sister is new root
                    elif sp.up is tre.treenode:
                        tre = [i for i in sp.up.children if i != sp][0]
                        tre = toytree.tree(tre)
                        tre.up = None


                    # if parent is a non-root node then connect sister to up.up
                    else:
                        # get sister
                        sister = [i for i in sp.up.children if i != sp][0]

                        # connect sister to grandparent
                        sister.up = sp.up.up

                        # drop parent from grandparent
                        sp.up.up.children.remove(sp.up)

                        # add sister to grandparent children
                        sp.up.up.children.append(sister)

                        # extend sisters dist to reach grandparent
                        sister.dist += sp.up.dist

                    ext += 1

            # update branch lengths so all tips end at time=present
            tips = tre.treenode.get_leaves()
            for x in tips:
                x.dist += dt

            # check stopping criterion
            if stop == "taxa":
                if len(tips) >= taxa_stop:
                    break
            else:
                if t >= time_stop:
                    break

        # report status
        if verbose:
            fill = (evnts - ext, ext, evnts / (evnts - ext), resets)
            print("b:\t{}\nd:\t{}\nb/d:\t{:.2f}\nreset:\t{}".format(*fill))

        # update coords and return
        tre.treenode.ladderize()
        tre._coords.update()

        # rename tips so names are in order else random
        nidx = list(range(tre.ntips))
        if random_names:
            random.shuffle(nidx)
        for idx, node in tre.idx_dict.items():
            if node.is_leaf():
                node.name = "r{}".format(nidx[idx])
        return tre


def _prune(tre, verbose=False):
    """
    Helper function for recursively pruning extinct branches in bd trees.
    Dynamic func!
    """
    ttree = tre.copy()
    tips = ttree.treenode.get_leaves()

    if np.any(np.array([x.height for x in tips]) > 0):
        for t in tips:
            if not np.isclose(t.height, 0):
                if verbose: 
                    print("Removing node/height {}/{}".format(t.name, t.height))
                t.delete(prevent_nondicotomic=False)
                ttree = _prune(ttree)
    return ttree


def return_small_clade(treenode):
    "used to produce balanced trees, returns a tip node from the smaller clade"
    node = treenode
    while 1:
        if node.children:
            c1, c2 = node.children
            node = sorted([c1, c2], key=lambda x: len(x.get_leaves()))[0]
        else:
            return node



#!/usr/bin/env python

"""
A custom Mark and mark generator to create Toytree drawings in toyplot.
"""
import numpy as np
import xml.etree.ElementTree as xml
import toyplot
from toyplot.mark import Mark
from toyplot.html import _draw_bar, _draw_triangle, _draw_circle, _draw_rect
from .TreeStyle import COLORS1
# from toytree.utils import ToytreeError

# Register multipledispatch to share with toyplot.html
import functools
from multipledispatch import dispatch
dispatch = functools.partial(dispatch, namespace=toyplot.html._namespace)

"""
- fixed-order extension to tip positions for missing labels..?
- rawtree consensus checking...
- container tree to Mark
"""


PATH_FORMAT = {
    'c': "M {px:.1f} {py:.1f} L {cx:.1f} {cy:.1f}",
    'b1': "M {px:.1f} {py:.1f} C {px:.1f} {cy:.1f}, {px:.1f} {cy:.1f}, {cx:.1f} {cy:.3f}",
    'b2': "M {px:.1f} {py:.1f} C {cx:.1f} {py:.1f}, {cx:.1f} {py:.1f}, {cx:.1f} {cy:.3f}",    
    'p1': "M {px:.1f} {py:.1f} L {px:.1f} {cy:.1f} L {cx:.1f} {cy:.1f}",
    'p2': "M {px:.1f} {py:.1f} L {cx:.1f} {py:.1f} L {cx:.1f} {cy:.1f}",
    'pc': "M {cx:.1f} {cy:.1f} L {dx:.1f} {dy:.1f} A {rr:.1f} {rr:.1f} 0 0 {flag} {px:.1f} {py:.1f}",
}


class ToytreeMark(Mark):
    """
    Custom mark testing.
    """
    def __init__(
        self, 
        ntable, 
        node_colors, 
        node_sizes, 
        node_style,
        node_markers,
        node_hover,
        etable,
        edge_colors,
        edge_widths,
        edge_type,
        edge_style,
        edge_align_style,
        tip_labels,
        tip_labels_angles,
        tip_labels_colors,
        tip_labels_style,
        node_labels,
        node_labels_style,
        tip_labels_align,
        layout,
        xbaseline,
        ybaseline,
        admixture_edges,
        shrink,
        **kwargs):

        # inherit type
        Mark.__init__(self)

        # check arg types
        self._coordinate_axes = ['x', 'y']  
        self.layout = layout
        self.shrink = (shrink if shrink else 0)

        # store anything here that you want available as tool-tip hovers
        self.ntable = ntable
        self.etable = etable

        # positioning
        self.xbaseline = xbaseline
        self.ybaseline = ybaseline
        self.ntable[:, 0] += xbaseline
        self.ntable[:, 1] += ybaseline

        # radial positioning
        self.radii = np.sqrt(
            self.ntable[:, 0] ** 2 + self.ntable[:, 1] ** 2 
        )

        # node plotting args
        self.node_colors = node_colors
        self.node_sizes = node_sizes
        self.node_markers = node_markers
        self.node_style = node_style
        self.node_hover = node_hover

        # edge (tree) plotting args
        self.edge_colors = edge_colors
        self.edge_widths = edge_widths
        self.edge_style = edge_style
        self.edge_type = edge_type
        self.edge_align_style = edge_align_style

        # tip labels
        self.tip_labels = tip_labels
        self.tip_labels_angles = tip_labels_angles
        self.tip_labels_colors = tip_labels_colors
        self.tip_labels_style = tip_labels_style
        self.tip_labels_align = tip_labels_align

        # node labels
        self.node_labels = node_labels
        self.node_labels_style = node_labels_style

        # admix edges
        self.admixture_edges = admixture_edges


    @property
    def nnodes(self):
        return self.ntable.shape[0]


    def domain(self, axis):
        """
        The domain of data that will be tracked.
        """
        index = self._coordinate_axes.index(axis)
        domain = toyplot.data.minimax(self.ntable[:, index])
        return domain


    def extents(self, axes):
        """
        Extends domain to fit tip names or mars based on their size, but 
        does not extend the data domain. 
        The main component to worry about here is tip labels text, especially
        for weird layouts that make it angled. For circular we project text
        at angle and inverted angle to get projection from anchor start,end.
        But node markers also contribute to this since extents=0 would cut a
        circular 15px node marker in half at the root. So we would need at 
        least 15px left extent.

        Further modification is made by the argument 'scale' which can
        compress the tree to allow more space for names. This is done by 
        making the extents larger.

        extents is [l, r, b, t] for each textbox
        """
        if self.layout != "c":
            # coordinates of all nodes on the tree.
            coords = (
                self.ntable[:, 0],
                self.ntable[:, 1],
            )

            # extents of all node markers
            ntips = len(self.tip_labels)
            nnodes = len(coords[0])

            # get tip label text extents
            if np.any(self.tip_labels):
                text_extents = toyplot.text.extents(
                    self.tip_labels,
                    self.tip_labels_angles,
                    style={
                        "-toyplot-vertical-align": "middle",
                        "font-family": "helvetica",
                        "font-weight": "normal",
                        "stroke": "none",
                        "font-size": toyplot.units.convert(
                            self.tip_labels_style['font-size'], "px") + 10,
                    }
                )

            else:
                text_extents = [[0, 0, 0, 0]] * ntips

            # check node extents
            xnode_sizes = self.node_sizes.copy()
            xnode_sizes[self.node_sizes == None] = 0

            # check edge widths
            # tmp fix for multitrees popping stroke-width currently
            try:
                xedge_widths = self.node_sizes.copy()
                xedge_widths[:-1] = self.edge_widths.copy()
                xedge_widths[xedge_widths == None] = self.edge_style['stroke-width']
            except KeyError:
                xedge_widths = np.repeat(2, self.nnodes)

            # empty extents for filling
            extents = (
                np.zeros(nnodes), np.zeros(nnodes),
                np.zeros(nnodes), np.zeros(nnodes),
            )

            # fill each node 
            for nidx in range(nnodes):

                # check tips extents 
                if nidx < ntips:
                    node_extent = np.array([xnode_sizes[nidx]] * 4)
                    node_extent *= [-1, 1, -1, 1]
                    edge_extent = np.array([xedge_widths[nidx]] * 4)
                    edge_extent *= [-1, 1, -1, 1]

                    if not np.any(self.tip_labels):
                        text_extent = np.array([0, 0, 0, 0])
                    else:
                        text_extent = [
                            text_extents[0][nidx],
                            text_extents[1][nidx],
                            text_extents[2][nidx],
                            text_extents[3][nidx],
                        ]

                        offset = toyplot.units.convert(
                            self.tip_labels_style["-toyplot-anchor-shift"],
                            "px",
                        )

                        # shrink extends the extents in the angle of the tips x layout
                        if self.layout == 'd':
                            width = abs(text_extent[2]) + abs(text_extent[3])
                            text_extent[3] = width + self.shrink + offset
                            text_extent[2] = 0

                        elif self.layout == 'u':
                            width = abs(text_extent[2]) + abs(text_extent[3])
                            text_extent[2] = -(width + offset + self.shrink)
                            text_extent[3] = 0

                        elif self.layout == 'r':
                            width = text_extent[1] - text_extent[0]
                            text_extent[0] = 0
                            text_extent[1] = width + offset * self.shrink

                        elif self.layout == 'l':
                            width = text_extent[1] - text_extent[0]
                            text_extent[0] = -(width + offset + self.shrink)
                            text_extent[1] = 0


                # check tips extents 
                else:
                    node_extent = np.array([xnode_sizes[nidx]] * 4)
                    node_extent *= [-1, 1, -1, 1]
                    edge_extent = np.array([xedge_widths[nidx]] * 4)
                    edge_extent *= [-1, 1, -1, 1]
                    text_extent = np.array([0, 0, 0, 0])

                # store extent nidx
                extents[0][nidx] = min([node_extent[0], edge_extent[0], text_extent[0]])
                extents[1][nidx] = max([node_extent[1], edge_extent[1], text_extent[1]])
                extents[2][nidx] = min([node_extent[2], edge_extent[2], text_extent[2]])
                extents[3][nidx] = max([node_extent[3], edge_extent[3], text_extent[3]])


        # for radial trees we want extents to fit similar in all directions
        # regardless of branch and tip name lengths. So find the radius of 
        # the circle + anchor shift + longest name and pass in all directions.
        else:
            coords = (
                max(self.radii) * np.array([-1, 0, 1, 0]),
                max(self.radii) * np.array([0, 1, 0, -1]),
            )

            # no tip labels for extends
            if all([i is None for i in self.tip_labels]):

                # no extents necessary
                if all([i is None for i in self.node_sizes]):
                    coords = tuple([np.array([])] * 2)
                    extents = tuple([np.array([])] * 4)

                # extend by node size                    
                else:
                    extents = (
                        np.repeat(max(self.node_sizes) * -1, 4),
                        np.repeat(max(self.node_sizes), 4),
                        np.repeat(max(self.node_sizes), 4),
                        np.repeat(max(self.node_sizes) * -1, 4),                                                                        
                    )

            # get the maxwidth of any tips ignoring positioning
            else:
                tips = self.tip_labels
                exts = toyplot.text.extents(
                    tips,
                    0, 
                    style={
                        "-toyplot-vertical-align": "middle",
                        "font-family": "helvetica",
                        "font-weight": "normal",
                        "stroke": "none",
                        "font-size": toyplot.units.convert(
                            self.tip_labels_style['font-size'], "px") + 10,
                    }
                )
                maxw = max([exts[1][i] - exts[0][i] for i in range(len(tips))])
                ashift = toyplot.units.convert(
                    self.tip_labels_style["-toyplot-anchor-shift"], "px")
                extents = (
                    np.repeat(-self.shrink - ashift - maxw * 1.5, 4),  # left ext
                    np.repeat(self.shrink + ashift + maxw * 1.5, 4),  # right ext
                    np.repeat(-self.shrink - ashift - maxw * 1.5, 4),  # bottom
                    np.repeat(self.shrink + ashift + maxw * 1.5, 4),  # top
                )

        # # all other layouts 
        # else:
        #     coords = (
        #         self.ntable[:len(self.tip_labels), 0],
        #         self.ntable[:len(self.tip_labels), 1],
        #     )
        #     style = {
        #         'font-size': toyplot.units.convert(self.tip_labels_style['font-size'], "px") + 5,
        #         '-toyplot-anchor-shift': self.tip_labels_style['-toyplot-anchor-shift']
        #     }
        #     extents = toyplot.text.extents(
        #         self.tip_labels,
        #         self.tip_labels_angles,
        #         style,
        #         #(self.tip_labels_style if self.tip_labels[0] else {}),
        #     )

        return coords, extents



class RenderToytree:
    """
    Organized class to call within _render
    """
    def __init__(self, axes, mark, context):

        # inputs
        self.mark = mark
        self.axes = axes
        self.context = context

        # to be constructed ...
        self.mark_xml = None

        # construction funcs
        self.project_coordinates()
        self.build_dom()


    def build_dom(self):
        """
        Creates DOM of xml.SubElements in self.context.
        """
        self.mark_toytree()
        self.mark_edges()
        self.mark_align_edges()
        self.mark_admixture_edges()
        self.mark_nodes()
        self.mark_node_labels()


        # for multitrees tips are sometimes not drawn.
        self.mark_tip_labels()


    def project_coordinates(self):
        """
        Stores node coordinates (data units) projecting as pixel units.
        """
        # project data coordinates into pixels
        self.nodes_x = self.axes.project('x', self.mark.ntable[:, 0])
        self.nodes_y = self.axes.project('y', self.mark.ntable[:, 1])
        if self.mark.layout == 'c':
            self.radii = self.axes.project('x', self.mark.radii)
            self.maxr = max(self.radii)

        # get align edge tips coords
        if self.mark.tip_labels_align:

            # coords of aligned tips across fixed x axis 0
            ntips = self.mark.tip_labels.size
            if self.mark.layout in ('r', 'l'):
                self.tips_x = np.repeat(
                    self.axes.project('x', self.mark.xbaseline), ntips)
                self.tips_y = self.nodes_y[:ntips]

            # coords of aligned tips across fixed y axis 0
            elif self.mark.layout in ('u', 'd'):
                self.tips_x = self.nodes_x[:ntips]
                self.tips_y = np.repeat(
                    self.axes.project('y', self.mark.ybaseline), ntips)

            # coords of tips around a circumference 
            elif self.mark.layout in ('c'):
                self.tips_x = np.zeros(ntips)
                self.tips_y = np.zeros(ntips)
                for idx, angle in enumerate(self.mark.tip_labels_angles):
                    radian = np.deg2rad(angle)
                    cx = 0 + max(self.mark.radii) * np.cos(radian)
                    cy = 0 - max(self.mark.radii) * np.sin(radian)
                    self.tips_x[idx] = self.axes.project('x', cx)
                    self.tips_y[idx] = self.axes.project('y', cy)             



    def get_paths(self):
        """
        # get edge table shape based on edge and layout types
        # 'c': M x y L x y                  # phylogram
        # 'p': M x y L x y L x y            # cladogram
        # 'b': M x y C x y, x y, x y        # bezier phylogram
        # 'f': M x y A r r, x, a, f, x y    # arcs/circle tree

        The arc/circle method applies to edge_type 'p' when layout='c'       
        """
        # modify order of x or y shift of edges for p,b types.
        if self.mark.edge_type in ('p', 'b'):
            if self.mark.layout == 'c':
                path_format = PATH_FORMAT["pc"]

            elif self.mark.layout in ('u', 'd'):
                path_format = PATH_FORMAT["{}2".format(self.mark.edge_type)]

            else:
                path_format = PATH_FORMAT["{}1".format(self.mark.edge_type)]
        else:
            path_format = PATH_FORMAT[self.mark.edge_type]

        # store paths here
        paths = []
        keys = []

        # countdown from root node idx to tips
        for eidx in range(self.mark.etable.shape[0] - 1, -1, -1):

            # get parent and child
            cidx = self.mark.etable[eidx, 1]
            pidx = self.mark.etable[eidx, 0]
            cx, cy = self.nodes_x[cidx], self.nodes_y[cidx]
            px, py = self.nodes_x[pidx], self.nodes_y[pidx]

            # get parent and child node angles from origin
            if self.mark.layout == 'c':
                ox = self.nodes_x[-1] + 0.000000123  # avoid cx == ox
                oy = self.nodes_y[-1] + 0.000000321
                pr = self.radii[pidx] - ox
                theta = np.arctan((oy - cy) / (cx - ox))               

                # trig to get hypotenuse from theta and parent radius
                if cx >= ox:
                    dx = ox + np.cos(theta) * pr
                    dy = oy - np.sin(theta) * pr
                else:
                    dx = ox - np.cos(theta) * pr
                    dy = oy + np.sin(theta) * pr

                # sweep flag to arc clockwise or not
                if py <= oy:
                    # this is working to arc within hemispheres
                    if dy <= oy:                          
                        if px >= dx:
                            flag = 1
                        else:
                            flag = 0
                    # if arc crosses the origin y
                    else:
                        flag = 1
                else:
                    # dy on same hemisphere?
                    if dy >= oy:
                        if px >= dx:
                            flag = 0
                        else:
                            flag = 1
                    else:
                        flag = 0

                # build paths.
                keys.append("{},{}".format(pidx, cidx))
                paths.append(
                    path_format.format(**{
                        'cx': cx, 'cy': cy, 'px': px, 'py': py,
                        'dx': dx, 'dy': dy, 'rr': pr, 'flag': flag, 
                    })
                )

            # build path string for simple types
            else: 
                keys.append("{},{}".format(pidx, cidx))                
                paths.append(
                    path_format.format(**{
                        'cx': cx, 'cy': cy, 'px': px, 'py': py,
                    })
                )
        return paths, keys



    def mark_toytree(self):
        """
        Creates the top-level Toytree mark.
        """
        self.mark_xml = xml.SubElement(
            self.context.parent, "g", 
            id=self.context.get_id(self.mark),
            attrib={"class": "toytree-mark-Toytree"},
        )



    def mark_edges(self):
        """
        Creates SVG paths for each tree edge under class toytree-Edges
        """
        # get paths based on edge type and layout
        paths, keys = self.get_paths()

        # get shared versus unique styles
        unique_styles = get_unique_edge_styles(self.mark)
        self.mark.edge_style['fill'] = "none"

        # render the edge group
        self.edges_xml = xml.SubElement(
            self.mark_xml, "g", 
            attrib={"class": "toytree-Edges"}, 
            style=style_to_string(self.mark.edge_style)
        )

        # render the edge paths
        for idx, path in enumerate(paths):
            style = unique_styles[idx]
            if style:
                xml.SubElement(
                    self.edges_xml, "path",
                    d=path,
                    id=keys[idx],
                    style=style_to_string(style),
                )
            else:
                xml.SubElement(
                    self.edges_xml, "path",
                    d=path,
                    id=keys[idx],
                )



    def mark_nodes(self):
        """
        Creates marker elements for each node under class toytree-Nodes.
        This could store ids to the nodes if we planned some interesting
        downstream JS interactivity...
        """
        # skip if all nodes are size=0
        if not all([i in (0, None) for i in self.mark.node_sizes]):

            # get fill style if differs among nodes
            unique_styles = [{} for i in range(self.mark.nnodes)]

            # only if variable tho
            if not all([i is None for i in self.mark.node_colors]):                

                # get fill and fill-opacity of each mark (levelorder)
                for idx in range(self.mark.nnodes):

                    # get the rgba node color
                    fill = self.mark.node_colors[idx]

                    # split into rgb and opacity and store result dict
                    unique_styles[idx] = split_rgba_style({'fill': fill})

            # Group all Nodes with shared style applied
            self.nodes_xml = xml.SubElement(
                self.mark_xml, "g", 
                attrib={"class": "toytree-Nodes"}, 
                style=style_to_string(self.mark.node_style),
            )

            # add node markers in reverse idx order (levelorder traversal)
            for nidx in range(self.mark.nnodes):

                # levelorder idx is root to tip idxs
                # idx = self.mark.nnodes - nidx

                # create marker with shape and size, e.g., <marker='o' size=12>
                marker = toyplot.marker.create(
                    shape=self.mark.node_markers[nidx],
                    size=self.mark.node_sizes[nidx],
                )

                # create the marker
                attrib = unique_styles[nidx]
                attrib['id'] = 'node-{}'.format(nidx)
                marker_xml = xml.SubElement(
                    self.nodes_xml, "g", attrib=attrib)

                # optionally add a title UNLESS node_label, then put the hover
                # on the node text instead.
                if self.mark.node_hover[nidx] is not None:
                    if self.mark.node_labels[nidx] is None:
                        xml.SubElement(marker_xml, "title").text = (
                            self.mark.node_hover[nidx])

                # project marker in coordinate space
                transform = "translate({:.3f},{:.3f})".format(
                    self.nodes_x[nidx],
                    self.nodes_y[nidx],
                )
                if marker.angle:
                    transform += " rotate({:.1f})".format(-marker.angle)
                marker_xml.set("transform", transform)

                # get shape type
                if marker.shape == "|":
                    _draw_bar(marker_xml, marker.size)
                elif marker.shape == "/":
                    _draw_bar(marker_xml, marker.size, angle=-45)
                elif marker.shape == "-":
                    _draw_bar(marker_xml, marker.size, angle=90)
                elif marker.shape == "\\":
                    _draw_bar(marker_xml, marker.size, angle=45)
                elif marker.shape == "+":
                    _draw_bar(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, angle=90)
                elif marker.shape == "x":
                    _draw_bar(marker_xml, marker.size, angle=-45)
                    _draw_bar(marker_xml, marker.size, angle=45)
                elif marker.shape == "*":
                    _draw_bar(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, angle=-60)
                    _draw_bar(marker_xml, marker.size, angle=60)
                elif marker.shape == "^":
                    _draw_triangle(marker_xml, marker.size)
                elif marker.shape == ">":
                    _draw_triangle(marker_xml, marker.size, angle=-90)
                elif marker.shape == "v":
                    _draw_triangle(marker_xml, marker.size, angle=180)
                elif marker.shape == "<":
                    _draw_triangle(marker_xml, marker.size, angle=90)
                elif marker.shape == "s":
                    _draw_rect(marker_xml, marker.size)
                elif marker.shape == "d":
                    _draw_rect(marker_xml, marker.size, angle=45)
                elif marker.shape and marker.shape[0] == "r":
                    width, height = marker.shape[1:].split("x")
                    _draw_rect(
                        marker_xml, marker.size,
                        width=float(width), height=float(height))
                elif marker.shape == "o":
                    _draw_circle(marker_xml, marker.size)
                elif marker.shape == "oo":
                    _draw_circle(marker_xml, marker.size)
                    _draw_circle(marker_xml, marker.size / 2)
                elif marker.shape == "o|":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size)
                elif marker.shape == "o/":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, -45)
                elif marker.shape == "o-":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, 90)
                elif marker.shape == "o\\":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, 45)
                elif marker.shape == "o+":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, 90)
                elif marker.shape == "ox":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, -45)
                    _draw_bar(marker_xml, marker.size, 45)
                elif marker.shape == "o*":
                    _draw_circle(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size)
                    _draw_bar(marker_xml, marker.size, -60)
                    _draw_bar(marker_xml, marker.size, 60)



    def mark_node_labels(self):
        """
        Creates text elements for node label under class toytree-NodeLabels.
        toytree-NodeLabels stores shared text styling but no positional style,
        and positional styling is interpreted and applied using transform
        methods in 'custom_draw_text' func, with unique text styling applied
        there (only 'fill' currently supported).
        """
        if not all([i is None for i in self.mark.node_labels]):

            # make xml with non-positioning styles that apply to all text
            style_group = {}
            exc = ["baseline-shift", "-toyplot-anchor-shift", "text-anchor"]
            style_pos = {"text-anchor": "middle", "stroke": "none"}
            style_pos.update(self.mark.node_labels_style)
            style_group = {
                i: j for (i, j) in style_pos.items() if i not in exc
            }

            # make the group with text style but not position styles
            nlabels_xml = xml.SubElement(
                self.mark_xml, "g", 
                attrib={"class": "toytree-NodeLabels"}, 
                style=toyplot.style.to_css(style_group),
            )

            # if title then put it here instead of on node marker
            for idx in range(self.mark.nnodes):

                label = self.mark.node_labels[idx]
                if label not in ("", " ", None):

                    # get size of text box based on style_pos
                    layout = toyplot.text.layout(
                        str(label),
                        style_pos,
                        toyplot.font.ReportlabLibrary(),
                    )

                    # apply transform to each textbox and add to xml
                    for child in layout.children:
                        for textbox in child.children:

                            # project points into coordinate space 
                            transform = "translate({:.2f},{:.2f})".format(
                                self.nodes_x[idx] + textbox.left,
                                self.nodes_y[idx] + textbox.baseline,
                            )
                            # if angle:
                            # transform += "rotate({:.1f})".format(-angle)

                            # create a group marker for positioning text
                            group = xml.SubElement(
                                nlabels_xml, "g", 
                            )
                            group.set("transform", transform)

                            # optionally add a title 
                            title = self.mark.node_hover[idx]
                            if title is not None:
                                xml.SubElement(group, "title").text = str(title)
                            xml.SubElement(group, "text").text = str(label)



    def mark_tip_labels(self):
        """
        Creates text elements for tip labels under class toytree-TipLabels.
        Styling here needs to support both linear, circular and unrooted trees,
        which is trick by combining style for -toyplot-anchor-shift and angles
        when setting transform.       
        """
        if not all([i is None for i in self.mark.tip_labels]):

            # allowed styling that can be updated by user style
            top_style = {
                'font-family': 'helvetica',
                'font-weight': 'normal',
                'white-space': 'pre',
                'fill': 'rgb(90.6%,54.1%,76.5%)',
                'fill-opacity': '1.0',
                'stroke': 'none',
                'font-size': '9px',
            }
            for sty in top_style:
                if sty in self.mark.tip_labels_style:
                    top_style[sty] = self.mark.tip_labels_style[sty]

            # apply font styling but NOT POSITIONAL styling to group.
            tips_xml = xml.SubElement(
                self.mark_xml, "g", 
                attrib={"class": "toytree-TipLabels"}, 
                style=style_to_string(top_style),
            )

            # add tip markers from 0 to ntips
            for tidx, tip in enumerate(self.mark.tip_labels):

                # allowed positional styling
                pos_style = {
                    "font-family": 'helvetica',
                    "font-size": "11px",
                    "-toyplot-anchor-shift": "15px",
                    "baseline-shift": "0px",
                    "text-anchor": "start",
                }
                for sty in pos_style:
                    if sty in self.mark.tip_labels_style:
                        pos_style[sty] = self.mark.tip_labels_style[sty]

                # get offset as int in case it needs to be flipped
                offset = toyplot.units.convert(
                    pos_style["-toyplot-anchor-shift"], "px",
                )

                # assign tip to class depending on coordinates
                if self.mark.layout == "r":
                    angle = self.mark.tip_labels_angles[tidx]
                    pos_style["text-anchor"] = "start"
                    pos_style["-toyplot-anchor-shift"] = offset

                elif self.mark.layout == "l":
                    angle = self.mark.tip_labels_angles[tidx]  # + 180
                    pos_style["text-anchor"] = "end"
                    pos_style["-toyplot-anchor-shift"] = -offset

                elif self.mark.layout == 'd':
                    angle = self.mark.tip_labels_angles[tidx]
                    pos_style["text-anchor"] = "end"
                    pos_style["-toyplot-anchor-shift"] = -offset

                elif self.mark.layout == "u":
                    angle = self.mark.tip_labels_angles[tidx]  # + 180
                    pos_style["text-anchor"] = "start"
                    pos_style["-toyplot-anchor-shift"] = offset

                elif self.mark.layout == "c":
                    angle = self.mark.tip_labels_angles[tidx]  # * -1
                    pos_style["text-anchor"] = "start"
                    pos_style["-toyplot-anchor-shift"] = offset
                    if (angle < -90) & (angle > -270):
                        angle += 180
                        pos_style["text-anchor"] = "end"
                        pos_style["-toyplot-anchor-shift"] = -offset

                # align tip at end for tip_labels_align=True
                cx = self.nodes_x[tidx]
                cy = self.nodes_y[tidx]
                if self.mark.tip_labels_align:
                    cx = self.tips_x[tidx]
                    cy = self.tips_y[tidx]

                # angle of text
                transform = "translate({:.2f},{:.2f})".format(cx, cy)
                transform += "rotate({:.0f})".format(angle)

                # the position of the tip TextBox
                tip_xml = xml.SubElement(tips_xml, "g")
                tip_xml.set("transform", transform)

                # split unique color if 
                icolor = self.mark.tip_labels_colors[tidx]
                colordict = {}
                if icolor is not None:
                    # try splitting color:
                    try:
                        colordict = split_rgba_style({"fill": icolor})
                    except Exception:
                        colordict = {"fill": icolor}

                # get baseline given font-size, etc.,
                layout = toyplot.text.layout(
                    tip,
                    pos_style,
                    toyplot.font.ReportlabLibrary(),
                )

                for line in layout.children:
                    for box in line.children:
                        xml.SubElement(
                            tip_xml,
                            "text", 
                            x="{:.2f}".format(box.left), 
                            y="{:.2f}".format(box.baseline),
                            style=style_to_string(colordict),
                            ).text = tip


    # def mark_tip_labels(self):
    #     """
    #     Creates text elements for tip labels under class toytree-TipLabels.
    #     Styling here needs to support both linear, circular and unrooted trees,
    #     which is trick by combining style for -toyplot-anchor-shift and angles
    #     when setting transform.       
    #     """
    #     if not all([i is None for i in self.mark.tip_labels]):

    #         # end anchor style updated for user text style
    #         top_style = {
    #             'font-family': 'helvetica',
    #             'font-weight': 'normal',
    #             'white-space': 'pre',
    #             'fill': 'rgb(90.6%,54.1%,76.5%)',
    #             'fill-opacity': '1.0',
    #             'stroke': 'none',
    #             'text-anchor': 'end',
    #             'font-size': '9px',
    #         }
    #         for sty in top_style:
    #             if sty in self.mark.tip_labels_style:
    #                 top_style[sty] = self.mark.tip_labels_style[sty]

    #         # force text-anchor to end of L-tips
    #         top_style['text-anchor'] = 'end'

    #         # apply font styling but NOT POSITIONAL styling to group.
    #         tips_left_xml = xml.SubElement(
    #             self.mark_xml, "g", 
    #             attrib={"class": "toytree-Tiplabels-L"}, 
    #             style=style_to_string(top_style),
    #         )

    #         # apply font styling but not position stylilng to group.
    #         top_style = {
    #             'font-weight': 'normal',
    #             'white-space': 'pre',
    #             'fill': 'rgb(90.6%,54.1%,76.5%)',
    #             'fill-opacity': '1.0',
    #             'stroke': 'none',
    #             'text-anchor': 'end',
    #             'font-size': '9px',
    #         }
    #         for sty in top_style:
    #             if sty in self.mark.tip_labels_style:
    #                 top_style[sty] = self.mark.tip_labels_style[sty]
    #         top_style['text-anchor'] = 'start'
    #         tips_right_xml = xml.SubElement(
    #             self.mark_xml, "g", 
    #             attrib={"class": "toytree-Tiplabels-R"}, 
    #             style=style_to_string(top_style)
    #         )

    #         # add tip markers
    #         for tidx, tip in enumerate(self.mark.tip_labels):

    #             icolor = self.mark.tip_labels_colors[tidx]
    #             tdict = {}
    #             if icolor is not None:
    #                 # try splitting color:
    #                 try:
    #                     tdict = split_rgba_style({"fill": icolor})
    #                 except Exception:
    #                     tdict = {"fill": icolor}

    #             # assign tip to class depending on coordinates
    #             if self.mark.layout == "r":
    #                 parent = tips_right_xml
    #                 angle = self.mark.tip_labels_angles[tidx]
    #             elif self.mark.layout == "l":
    #                 parent = tips_left_xml
    #                 angle = self.mark.tip_labels_angles[tidx]
    #             elif self.mark.layout == "d":
    #                 parent = tips_left_xml
    #                 angle = self.mark.tip_labels_angles[tidx] + 90
    #             elif self.mark.layout == "u":
    #                 parent = tips_right_xml
    #                 angle = self.mark.tip_labels_angles[tidx] - 90
    #             elif self.mark.layout == "c":
    #                 angle = self.mark.tip_labels_angles[tidx]
    #                 if (angle > 90) and (angle < 270):
    #                     parent = tips_left_xml
    #                 else:
    #                     parent = tips_right_xml

    #             # short variables
    #             cx = self.nodes_x[tidx] 
    #             cy = self.nodes_y[tidx]
    #             style_pos = self.mark.tip_labels_style
    #             style_text = tdict
    #             title = None

    #             # align tip at end for tip_labels_align=True
    #             if self.mark.tip_labels_align:
    #                 cx = self.tips_x[tidx]
    #                 cy = self.tips_y[tidx]

    #             # get baseline given font-size, etc.,
    #             layout = toyplot.text.layout(
    #                 tip,
    #                 style_text,
    #                 toyplot.font.ReportlabLibrary(),
    #             )
    #             baseline = layout.children[0].children[0].baseline

    #             # adjust projections based on angle and shift args
    #             ashift = toyplot.units.convert(style_pos["-toyplot-anchor-shift"], "px")

    #             # if right facing then use anchor-shift to +x, else -x
    #             if parent.attrib['class'] == "toytree-Tiplabels-R":

    #                 # anchor shifts left, baseline corrects y
    #                 if self.mark.layout == "r":
    #                     cx += ashift
    #                     cy += baseline

    #                 # anchor shifts up, baseline shifts right, angle is 90
    #                 elif self.mark.layout == "u":
    #                     cx += baseline
    #                     cy -= ashift
    #                     angle += 90

    #                 # 
    #                 elif self.mark.layout == 'c':
    #                     # convert ashift to the angle
    #                     if not angle:
    #                         cx += ashift
    #                         cy += baseline
    #                     else:
    #                         # get lengths from trig.
    #                         trans = toyplot.transform.rotation(angle)[0]                
    #                         ashift_x = ashift * trans[0, 0]
    #                         ashift_y = ashift * trans[0, 1]
    #                         bshift_y = baseline * trans[0, 0]
    #                         bshift_x = baseline * trans[0, 1]
    #                         cx += ashift_x + bshift_x
    #                         cy -= ashift_y - bshift_y

    #             else:

    #                 # anchor shifts down, baseline shifts left, angle is -90
    #                 if self.mark.layout == "d":
    #                     cx += baseline
    #                     cy += ashift
    #                     angle += 90

    #                 elif self.mark.layout == 'l':
    #                     cx -= ashift
    #                     cy += baseline
    #                     angle += 180

    #                 elif self.mark.layout == 'c':
    #                     angle += 180

    #                     # get lengths from trig.
    #                     trans = toyplot.transform.rotation(angle)[0]                
    #                     ashift_x = ashift * trans[0, 0]
    #                     ashift_y = ashift * trans[0, 1]
    #                     bshift_y = baseline * trans[0, 0]
    #                     bshift_x = baseline * trans[0, 1]
    #                     cx = cx - ashift_x + bshift_x
    #                     cy = cy + ashift_y + bshift_y

    #             # project points into coordinate space 
    #             transform = "translate({:.2f},{:.2f})".format(cx, cy)
    #             if angle:
    #                 transform += "rotate({:.1f})".format(-angle)

    #             # create a group marker for positioning text
    #             if style_text:
    #                 group = xml.SubElement(
    #                     parent, "g", style=style_to_string(style_text))
    #             else:
    #                 group = xml.SubElement(parent, "g")
    #             group.set("transform", transform)

    #             # optionally add a title 
    #             if title is not None:
    #                 xml.SubElement(group, "title").text = str(title)

    #             # style text should only include unique styling which currently for 
    #             # nodes is nothing, and for tips is only 'fill' and 'fill-opacity'.
    #             xml.SubElement(group, "text").text = tip




    def mark_align_edges(self):
        """
        Creates SVG paths for from each tip to 0 or radius.
        """
        # get paths based on edge type and layout
        if self.mark.tip_labels_align:
            apaths = []
            for tidx in range(len(self.mark.tip_labels)):

                adict = {
                    'cx': self.nodes_x[tidx],
                    'cy': self.nodes_y[tidx],
                    'px': self.tips_x[tidx], 
                    'py': self.tips_y[tidx],
                }
                path = PATH_FORMAT['c'].format(**adict)
                apaths.append(path)

            # render the edge group
            self.align_xml = xml.SubElement(
                self.mark_xml, "g", 
                attrib={"class": "toytree-AlignEdges"}, 
                style=style_to_string(self.mark.edge_align_style),
            )

            # render the edge paths
            for idx, path in enumerate(apaths):
                xml.SubElement(
                    self.align_xml, "path",
                    d=path,
                    # style=None,
                )



    def mark_admixture_edges(self):
        """
        Creates an SVG path for an admixture edge. The edge takes the same
        style as the edge_type of the tree.
        """
        if self.mark.admixture_edges is None:
            return 

        # iterate over colors for subsequent edges unless provided
        DEFAULT_ADMIXTURE_EDGES_STYLE = {
            "stroke": COLORS1[3], 
            "stroke-width": 5, 
            "stroke-opacity": 0.6,
            "stroke-linecap": "round", 
            "fill": "none",
            "font-size": "14px"
        }

        # create edge group element
        self.admix_xml = xml.SubElement(
            self.mark_xml, 'g',
            attrib={'class': 'toytree-AdmixEdges'},
            style=style_to_string(DEFAULT_ADMIXTURE_EDGES_STYLE)
        )

        # get position of 15% tipward from source point
        PATH = [
            "M {sdx:.1f} {sdy:.1f}",
            "L {sux:.1f} {suy:.1f}",
            "L {ddx:.1f} {ddy:.1f}",
            "L {dux:.1f} {duy:.1f}",
        ]            

        # ensure admixture_edges is a list of tuples
        if not isinstance(self.mark.admixture_edges, list):
            self.mark.admixture_edges = [self.mark.admixture_edges]

        # drwa each edge
        for aedge in self.mark.admixture_edges:

            # check if nodes have an overlapping interval
            src, dest, aprop, estyle, label = aedge

            # get their parents coord positions
            try:
                ps = self.mark.etable[self.mark.etable[:, 1] == src, 0][0]
                pd = self.mark.etable[self.mark.etable[:, 1] == dest, 0][0]

            # except if root edge
            except IndexError:
                raise NotImplementedError(
                    "whoops, admixture edge with root node. TODO.")

            # shared midpoint or separate midpoint (if edges do not overlap)
            # then only separate is possible).
            shared = False
            if isinstance(aprop, (int, float)):
                shared = True
                aprop = (aprop, aprop)

            # separate for each layout b/c its haaaard.
            if self.mark.layout == "r":
                sx, sy = self.nodes_y[src], self.nodes_x[src]
                dx, dy = self.nodes_y[dest], self.nodes_x[dest]
                psx, psy = self.nodes_y[ps], self.nodes_x[ps]
                pdx, pdy = self.nodes_y[pd], self.nodes_x[pd]

                disjoint = (psy >= dy) or (sy <= pdy)
                if (disjoint) or (not shared):
                    src_mid_y = sy - (abs(sy - psy) * aprop[0])
                    dest_mid_y = dy - (abs(dy - pdy) * aprop[1])
                else:
                    # get height of the admix line at midshared.
                    amin = min([sy, dy])
                    amax = max([psy, pdy])
                    admix_ymid = amin + (amax - amin) * aprop[0]
                    dest_mid_y = src_mid_y = admix_ymid

            elif self.mark.layout == "l":
                sx, sy = self.nodes_y[src], self.nodes_x[src]
                dx, dy = self.nodes_y[dest], self.nodes_x[dest]
                psx, psy = self.nodes_y[ps], self.nodes_x[ps]
                pdx, pdy = self.nodes_y[pd], self.nodes_x[pd]

                disjoint = (psy <= dy) or (sy >= pdy)
                if disjoint or (not shared):
                    src_mid_y = sy + (abs(sy - psy) * aprop[0])
                    dest_mid_y = dy + (abs(dy - pdy) * aprop[1])
                else:
                    # get height of the admix line at midshared.
                    amin = max([sy, dy])
                    amax = min([psy, pdy])
                    admix_ymid = amin + abs(amax - amin) * aprop[0]
                    dest_mid_y = src_mid_y = admix_ymid

            elif self.mark.layout == "d":
                sx, sy = self.nodes_x[src], self.nodes_y[src]
                dx, dy = self.nodes_x[dest], self.nodes_y[dest]
                psx, psy = self.nodes_x[ps], self.nodes_y[ps]
                pdx, pdy = self.nodes_x[pd], self.nodes_y[pd]

                disjoint = (psy >= dy) or (sy <= pdy)
                if disjoint or (not shared):
                    src_mid_y = sy - (abs(sy - psy) * aprop[0])
                    dest_mid_y = dy - (abs(dy - pdy) * aprop[1])
                else:
                    # get height of the admix line at midshared.
                    amin = min([sy, dy])
                    amax = max([psy, pdy])
                    admix_ymid = amin - abs(amax - amin) * aprop[0]
                    dest_mid_y = src_mid_y = admix_ymid                   
                    # if aprop[0] == aprop[1]:
                    #     admix_ymid = amin + (amax - amin) * aprop[0]
                    #     dest_mid_y = src_mid_y = admix_ymid
                    # else:
                    #     src_mid_y = sy - (abs(sy - psy) * aprop[0])
                    #     dest_mid_y = dy - (abs(dy - pdy) * aprop[1])

            elif self.mark.layout == "u":
                sx, sy = self.nodes_x[src], self.nodes_y[src]
                dx, dy = self.nodes_x[dest], self.nodes_y[dest]
                psx, psy = self.nodes_x[ps], self.nodes_y[ps]
                pdx, pdy = self.nodes_x[pd], self.nodes_y[pd]

                disjoint = (psy <= dy) or (sy >= pdy)
                if disjoint or (not shared):
                    src_mid_y = sy + (abs(sy - psy) * aprop[0])
                    dest_mid_y = dy + (abs(dy - pdy) * aprop[1])
                else:
                    # get height of the admix line at midshared.
                    amin = max([sy, dy])
                    amax = min([psy, pdy])
                    admix_ymid = amin + (abs(amax - amin) * aprop[0])
                    dest_mid_y = src_mid_y = admix_ymid


            # project angle of up/down lines towards parent nodes.
            if self.mark.edge_type == "c":            

                # angle from src to src parent
                if (psx - sx) == 0:
                    x_shift_src_mid = 0
                else:
                    theta = np.arctan((psy - sy) / (psx - sx))
                    x_shift_src_mid = (src_mid_y - sy) / np.tan(theta)                

                # angle from dest to dest parent
                if (pdx - dx) == 0:
                    x_shift_dest_mid = 0
                else:
                    theta = np.arctan((pdy - dy) / (pdx - dx))
                    x_shift_dest_mid = (dest_mid_y - dy) / np.tan(theta)                
                xend = pdx

            else:
                x_shift_dest_mid = 0
                x_shift_src_mid = 0
                xend = dx

            # build the SVG path
            if self.mark.layout in ("r", "l"):
                edge_dict = {
                    'sdy': sx,  # + x_shift_src_tip + snudge,
                    'sdx': sy,  # src_tip_y, 
                    'suy': sx + x_shift_src_mid,
                    'sux': src_mid_y,  # admix_ymid,
                    'ddy': dx + x_shift_dest_mid,
                    'ddx': dest_mid_y,  # admix_ymid,
                    'duy': xend,
                    'dux': pdy,  # dest_tip_y,
                }
                # tri_dict = {
                #     'x0': admix_ymid - 6,
                #     'x1': admix_ymid + 6,
                #     'x2': admix_ymid,
                #     'y0': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'y1': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'y2': np.mean([edge_dict['suy'], edge_dict['ddy']]) + 8,
                # }

            else:            
                edge_dict = {
                    'sdx': sx,             # + x_shift_src_tip + snudge,
                    'sdy': sy,  # src_tip_y, 

                    'sux': sx + x_shift_src_mid,
                    'suy': src_mid_y,  # admix_ymid,

                    'ddx': dx + x_shift_dest_mid,
                    'ddy': dest_mid_y,  # admix_ymid,

                    'dux': xend,
                    'duy': pdy,  # dest_tip_y,
                }

                # TODO: not finished aligning triangle/arrow
                # tri_dict = {
                #     'y0': admix_ymid - 6,
                #     'y1': admix_ymid + 6,
                #     'y2': admix_ymid,
                #     'x0': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'x1': np.mean([edge_dict['suy'], edge_dict['ddy']]) - 6,
                #     'x2': np.mean([edge_dict['suy'], edge_dict['ddy']]) + 8,
                # }

            # EDGE path                
            path = " ".join(PATH).format(**edge_dict)
            xml.SubElement(
                self.admix_xml, "path",
                d=path,
                style=style_to_string(estyle),
            )

            lstyle = estyle.copy()
            # LABEL
            if label is not None:

                # RENDER edge label
                lstyle['fill'] = '#262626'
                lstyle['fill-opacity'] = '1.0'
                lstyle['stroke'] = "none"
                lstyle['text-anchor'] = 'middle'

                # position
                if self.mark.layout in ("r", "l"):
                    xtext = np.mean([sx + x_shift_src_mid, dx + x_shift_dest_mid])
                    ytext = np.mean([src_mid_y, dest_mid_y])
                    xtext += 12
                else:
                    ytext = np.mean([sx + x_shift_src_mid, dx + x_shift_dest_mid])
                    xtext = np.mean([src_mid_y, dest_mid_y])

                xml.SubElement(
                    self.admix_xml,
                    "text",
                    x="{:.2f}".format(ytext),
                    y="{:.2f}".format(xtext),
                    style=style_to_string(lstyle),
                ).text = str(label)


                # # allowed styling that can be updated by user style
                # top_style = {
                #     'font-family': 'helvetica',
                #     'font-weight': 'normal',
                #     'white-space': 'pre',
                #     'fill': estyle['stroke'],
                #     'fill-opacity': '1.0',
                #     'stroke': 'none',
                #     'font-size': '9px',
                #     'baseline-shift': '0px',                
                # }

                # # apply font styling but NOT POSITIONAL styling to group.
                # lab_xml = xml.SubElement(
                #     self.admix_xml, "g", 
                #     attrib={"class": "toytree-AdmixEdge-Label"}, 
                #     style=style_to_string(top_style),
                # )

                # # add tip markers from 0 to ntips
                # for tidx, tip in enumerate(self.mark.tip_labels):

                #     # angle of text
                #     transform = "translate({:.2f},{:.2f})".format(xtext, ytext)

                #     # the position of the tip TextBox
                #     t_xml = xml.SubElement(lab_xml, "g")
                #     t_xml.set("transform", transform)

                #     # get baseline given font-size, etc.,
                #     layout = toyplot.text.layout(
                #         tip,
                #         top_style,
                #         toyplot.font.ReportlabLibrary(),
                #     )

                #     for line in layout.children:
                #         for box in line.children:
                #             xml.SubElement(
                #                 t_xml,
                #                 "text", 
                #                 x="{:.2f}".format(box.left), 
                #                 y="{:.2f}".format(box.baseline),
                #                 style={"fill": "green"},
                #                 # style=style_to_string(colordict),
                #                 ).text = tip






            # RENDER TRIANGLE
            # xml.SubElement(
            #     self.admix_xml, "polygon",
            #     points=(
            #         "{:.0f},{:.0f} {:.0f},{:.0f} {:.0f},{:.0f}"
            #         .format(
            #             tri_dict['x0'], tri_dict['y0'],
            #             tri_dict['x1'], tri_dict['y1'],
            #             tri_dict['x2'], tri_dict['y2'],
            #         )
            #     ),
            #     attrib={
            #         "fill": self.mark.admixture_edges_style["stroke"],
            #         "stroke-dasharray": "0,0",
            #     },
            # )
            # markup.set("transform", "rotate({})".format(-angle))


        # _draw_triangle(marker_xml, marker.size, angle=90)
        # def _draw_triangle(parent_xml, size, angle=0):
        # markup = xml.SubElement(
        #     parent_xml,
        #     "polygon",
        #     points=" ".join(["%r,%r" % (xp, yp) for xp, yp in [
        #        (-size / 2, size / 2),
        #        (0, -size / 2),
        #        (size / 2, size / 2),
        #        ]]),
        #     )
        # if angle:
        #     markup.set("transform", "rotate(%r)" % (-angle,))




@dispatch(toyplot.coordinates.Cartesian, ToytreeMark, toyplot.html.RenderContext)
def _render(axes, mark, context):
    RenderToytree(axes, mark, context)





# HELPER FUNCTIONS ----------------------




def get_unique_edge_styles(mark):
    """
    Reduces node styles to prevent redundancy in HTML.
    """
    # minimum styling of node markers
    unique_styles = [{} for i in range(mark.etable.shape[0])]

    # if edge widths and edge colors are both empty then just return
    check0 = all([i is None for i in mark.edge_colors])
    check1 = all([i is None for i in mark.edge_widths])
    if check0 & check1:
        return unique_styles

    # iterate through styled node marks to get shared styles and expand axes
    for idx in range(mark.etable.shape[0]):

        if mark.edge_widths[idx] is not None:
            unique_styles[idx]['stroke-width'] = mark.edge_widths[idx]

        if mark.edge_colors[idx] is not None:
            subd = split_rgba_style({'stroke': mark.edge_colors[idx]})
            unique_styles[idx]['stroke'] = subd['stroke']
            if subd['stroke-opacity'] != mark.edge_style["stroke-opacity"]:
                unique_styles[idx]['stroke-opacity'] = subd['stroke-opacity']

    return unique_styles




def split_rgba_style(style):
    """
    Because many applications (Inkscape, Adobe Illustrator, Qt) don't handle 
    CSS rgba() colors correctly this function does a work-around.
    Takes a CSS color in rgba, e.g., 'rgba(40.0%,76.1%,64.7%,1.000)' 
    labeled in a dictionary as {'fill': x, 'fill-opacity': y} and 
    returns with fill as rgb and fill-opacity from rgba or clobbered
    by the fill-opacity arg. Similar functionality for stroke, stroke-opacity.
    """
    if "fill" in style:
        color = style['fill']
        try:
            color = toyplot.color.css(color)
        except (TypeError, AttributeError):
            # print(type(color), color)
            pass

        if str(color) == "none":
            style["fill"] = "none"
            style["fill-opacity"] = 1.0
        else:
            rgb = "rgb({:.3g}%,{:.3g}%,{:.3g}%)".format(
                color["r"] * 100, 
                color["g"] * 100, 
                color["b"] * 100,
            )
            style["fill"] = rgb
            style["fill-opacity"] = str(color["a"])


    if "stroke" in style:
        color = style['stroke']
        try:
            color = toyplot.color.css(color)
        except (TypeError, AttributeError):
            # print(type(color), color)            
            pass

        if str(color) == "none":
            style["stroke"] = "none"
            style["stroke-opacity"] = 1.0
        else:
            rgb = "rgb({:.3g}%,{:.3g}%,{:.3g}%)".format(
                color["r"] * 100, 
                color["g"] * 100, 
                color["b"] * 100,
            )
            style["stroke"] = rgb
            style["stroke-opacity"] = str(color["a"])
    return style




def style_to_string(style):
    """
    Takes a style dict and writes to ordered style text:     
    input: {'fill': 'rgb(100%,0%,0%', 'fill-opacity': 1.0}
    returns: 'fill:rgb(100%,0%,0%);fill-opacity:1.0'
    """
    strs = ["{}:{}".format(key, value) for key, value in sorted(style.items())]
    return ";".join(strs)





# def check_admix_tuple_args(args):
#     """
#     Split list into bits.
#     """
#     # check that the argument is a list
#     assert isinstance(args, list), (
#         "admixture edges should be a list of tuples")

#     # iterate over the tuples
#     admixture_edges = []
#     for tup in args:

#         # parse required args
#         source = tup[0]
#         dest = tup[1]

#         # parse optional args
#         try:
#             height = tup[2]
#         except IndexError:
#             height = None

#         try:
#             width = (tup[3] if tup[3] else 3)
#         except IndexError:
#             width = 3

#         try:
#             color = (tup[4] if tup[4] else next(self.colors))
#         except IndexError:
#             color = next(self.colors)

#         try:
#             offset = (tup[5] if tup[5] else 0.05)
#         except IndexError:
#             offset = 0.05

#         # put it back into list as full tuples
#         admixture_edges.append(
#             (source, dest, height, style)



# # TODO
# def _render_text_file(owner, key, label, table, filename, context):
#     """
#     A variant of toyplot.html._render_table that can be used instead
#     to download a text file.
#     """
#     if isinstance(owner, toyplot.mark.Mark) and owner.annotation:
#         return
#     if isinstance(owner, toyplot.coordinates.Table) and owner.annotation:
#         return

#     names = []
#     columns = []

#     if isinstance(table, toyplot.data.Table):
#         for name, column in table.items():
#             if "toyplot:exportable" in table.metadata(name) and table.metadata(name)["toyplot:exportable"]:
#                 if column.dtype == toyplot.color.dtype:
#                     raise ValueError("Color column table export isn't supported.") # pragma: no cover
#                 else:
#                     names.append(name)
#                     columns.append(column.tolist())
#     else: # Assume numpy matrix
#         for column in table.T:
#             names.append(column[0])
#             columns.append(column[1:].tolist())

#     if not (names and columns):
#         return

#     owner_id = context.get_id(owner)
#     if filename is None:
#         filename = "toyplot"

#     context.require(
#         dependencies=["toyplot/menus/context", "toyplot/io"],
#         arguments=[owner_id, key, label, names, columns, filename],
#         code="""function(tables, context_menu, io, owner_id, key, label, names, columns, filename)
#         {
#             tables.set(owner_id, key, names, columns);

#             var owner = document.querySelector("#" + owner_id);
#             function show_item(e)
#             {
#                 return owner.contains(e.target);
#             }

#             function choose_item()
#             {
#                 io.save_file("text/csv", "utf-8", tables.get_csv(owner_id, key), filename + ".csv");
#             }

#             context_menu.add_item("Save " + label + " as CSV", show_item, choose_item);
#         }""",
#     )



#!/usr/bin/env python

from __future__ import print_function, division, absolute_import


import re
import os
from copy import deepcopy
import numpy as np
import toytree
import toyplot


#######################################################
# Exception Classes
#######################################################
class ToytreeError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class TreeError(Exception):
    "A problem occurred during a TreeNode operation"
    def __init__(self, value=''):
        self.value = value

    def __str__(self):
        return repr(self.value)


# TREE FORMATS
NW_FORMAT = {
    # flexible with support
    # Format 0 = (A:0.35,(B:0.72,(D:0.60,G:0.12)1.00:0.64)1.00:0.56);
    0: [
        ('name', str, True),
        ('dist', float, True),
        ('support', float, True),
        ('dist', float, True),
    ],

    # flexible with internal node names
    # Format 1 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E:0.64)C:0.56);
    1: [
        ('name', str, True),
        ('dist', float, True),
        ('name', str, True),
        ('dist', float, True),      
    ],

    # strict with support values
    # Format 2 = (A:0.35,(B:0.72,(D:0.60,G:0.12)1.00:0.64)1.00:0.56);
    2: [
        ('name', str, False),
        ('dist', float, False),
        ('support', str, False),
        ('dist', float, False),      
    ],

    # strict with internal node names
    # Format 3 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E:0.64)C:0.56);
    3: [
        ('name', str, False),
        ('dist', float, False),
        ('name', str, False),
        ('dist', float, False),      
    ],

    # strict with internal node names
    # Format 4 = (A:0.35,(B:0.72,(D:0.60,G:0.12)));
    4: [
        ('name', str, False),
        ('dist', float, False),
        (None, None, False),
        (None, None, False),      
    ],

    # Format 5 = (A:0.35,(B:0.72,(D:0.60,G:0.12):0.64):0.56);
    5: [
        ('name', str, False),
        ('dist', float, False),
        (None, None, False),
        ('dist', float, False),      
    ],

    # Format 6 = (A:0.35,(B:0.72,(D:0.60,G:0.12)E)C);
    6: [
        ('name', str, False),
        (None, None, False),
        (None, None, False),        
        ('dist', float, False),      
    ],

    # Format 7 = (A,(B,(D,G)E)C);
    7: [
        ('name', str, False),
        ('dist', float, False),
        ('name', str, False),        
        (None, None, False),        
    ],    


    # Format 8 = (A,(B,(D,G)));
    8: [
        ('name', str, False),
        (None, None, False),
        ('name', str, False),        
        (None, None, False),
    ],

    # Format 9 = (,(,(,)));
    9: [
        ('name', str, False),
        (None, None, False),
        (None, None, False),
        (None, None, False),
    ],    

    # Format 10 = ((a[&Z=1,Y=2]:1.0[&X=3], b[&Z=1,Y=2]:3.0[&X=2]):1.0[&L=1,W=0], ...
    # NHX Like mrbayes NEXUS common
    10: [
        ('name', str, True),
        ('dist', str, True),
        ('name', str, True),
        ('dist', str, True),
    ]
}



def parse_network(net, disconnect=True):
    """
    Parse network to extract the major topology. 
    This leaves the hybrid nodes in the tree and labels each with 
    .name="H{int}" and .gamma={float}.
    """
    # if net is a file then read the first line
    if os.path.exists(net):
        with open(net, 'r') as infile:
            net = infile.readline()

    # trim off loglik and anything after it (TODO: keep loglik)
    if ";" in net:
        net = net.split(";")[0] + ';'

    # sub :xxx:: to be ::: b/c I don't care about admix edge bls
    net = re.sub(r":\d.\w*::", ":::", net)

    # change H nodes to proper format
    while ",#" in net:
        pre, post = net.split(",#", 1)
        npre, npost = post.split(")", 1)
        newpre = npre.split(":")[0] + "-" + npre.split(":")[-1]
        net = pre + ")#" + newpre + npost
    net = net.replace(":::", "-")

    # parse cleaned newick and set empty gamma on all nodes
    net = toytree.tree(net, tree_format=1)

    # store admix data
    admix = {}

    # if not rooted choose any non-H root
    if not net.is_rooted():
        net = net.root(
            [i for i in net.get_tip_labels() if not i.startswith("#H")][0]
        )

    # Traverse tree to find hybrid nodes. If a hybrid node is labeled as a 
    # distinct branch in the tree then it is dropped from the tree and 
    for node in net.treenode.traverse("postorder"):

        # find hybrid nodes as internal nchild=1, or external with H in name
        if (len(node.children) == 1) or node.name.startswith("#H"):

            # assign name and gamma to hybrid nodes
            aname, aprop = node.name.split("-")
            aname = aname.lstrip("#")
            node.name = aname

            # assign hybrid to closest nodes up and down from edge
            # node.children[0].hybrid = int(aname[1:])
            # node.gamma = round(float(aprop), 3)
            # node.up.hybrid = int(aname[1:])

            # if root is a hybrid edge (ugh)
            if node.up is None:
                small, big = sorted(node.children, key=lambda x: len(x))
                root = toytree.TreeNode.TreeNode(name='root')
                node.children = [small]
                small.up = node
                node.up = root
                big.up = root
                root.children = [node, big]
                net.treenode = root

            # disconnect node by connecting children to parent
            if disconnect:

                # if tip is a hybrid
                if not node.children:
                    # get sister node
                    sister = [i for i in node.up.children if i != node][0]

                    # connect sister to gparent
                    sister.up = node.up.up
                    node.up.up.children.remove(node.up)
                    node.up.up.children.append(sister)

                # if hybrid is internal
                else:
                    node.up.children.remove(node)
                    for child in node.children:
                        child.up = node.up
                        node.up.children.append(child)

            # store admix data by descendants but remove hybrid tips
            desc = node.get_leaf_names()
            if aname in desc:
                desc = [i for i in node.up.get_leaf_names() if i != aname]
            desc = [i for i in desc if not i.startswith("#H")]

            # put this node into admix
            if aname not in admix:
                admix[aname] = (desc, aprop)

            # matching edge in admix, no arrange into correct order by minor
            else:
                # this is the minor edge
                if aprop < admix[aname][1]:
                    admix[aname] = (
                        admix[aname][0], 
                        desc, 
                        0.5, 
                        {}, 
                        str(round(float(aprop), 3)),
                    )

                # this is the major edge
                else:
                    admix[aname] = (
                        desc, 
                        admix[aname][0], 
                        0.5, 
                        {}, 
                        str(round(float(admix[aname][1]), 3)),
                    )

    # update coords needed if node disconnection is turned back on.
    net._coords.update()
    net = net.ladderize()
    return net, admix




class Annotator(object):
    """
    Add annotations as a new mark on top of an existing toytree mark.
    """
    def __init__(self, tree, axes, mark):
        self.tree = tree
        self.axes = axes
        self.mark = mark


    def draw_clade_box(
        self, 
        names=None, 
        regex=None, 
        wildcard=None, 
        yspace=None, 
        xspace=None, 
        **kwargs):
        """
        Draw a rectangle around a clade on a toytree.

        Parameters:
        -----------
        names, regex, wildcard:
            Choose one of these three methods to select one or more tipnames. 
            The clade composing all descendants of their common ancestor will 
            be highlighted.

        yspace (float or None):
            The extent to which boxes extend above and below the root and tip
            nodes. If None then this is automatically generated.

        xspace (float or None):
            The extent to which the clade box extends to the sides 
            (out of the clade towards other tips.) If None default uses 0.5.

        kwargs:
            Additional styling options are supported: color, opacity, etc.

        Returns:
        ------------
        Toyplot.mark.Range
        """

        # get the common ancestor
        nidx = self.tree.get_mrca_idx_from_tip_labels(
            names=names, regex=regex, wildcard=wildcard)

        # get tips descended from mrca
        tips = self.tree.idx_dict[nidx].get_leaves()
        tidxs = [i.idx for i in tips]

        # extent to which box bounds extend outside of the exact clade size.
        if not yspace:
            yspace = self.tree.treenode.height / 15.
        if not xspace:
            xspace = 0.45

        # left and right positions
        if self.mark.layout == 'r':
            xmin = self.mark.ntable[nidx, 0] - yspace
            xmax = max(self.mark.ntable[tidxs, 0]) + yspace
            ymin = min(self.mark.ntable[tidxs, 1]) - xspace
            ymax = max(self.mark.ntable[tidxs, 1]) + xspace   

        if self.mark.layout == 'l':
            xmin = self.mark.ntable[nidx, 0] + yspace
            xmax = max(self.mark.ntable[tidxs, 0]) - yspace
            ymin = max(self.mark.ntable[tidxs, 1]) + xspace
            ymax = min(self.mark.ntable[tidxs, 1]) - xspace   

        elif self.mark.layout == 'd':
            ymax = self.mark.ntable[nidx, 1] + yspace
            ymin = min(self.mark.ntable[tidxs, 1]) - yspace
            xmin = min(self.mark.ntable[tidxs, 0]) - xspace
            xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

        elif self.mark.layout == 'u':
            ymin = self.mark.ntable[nidx, 1] - yspace
            ymax = min(self.mark.ntable[tidxs, 1]) + yspace
            xmin = min(self.mark.ntable[tidxs, 0]) - xspace
            xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

        # draw the rectangle
        newmark = self.axes.rectangle(xmin, xmax, ymin, ymax, **kwargs)

        # put tree at the top of the scenegraph
        self.axes._scenegraph.remove_edge(self.axes, 'render', self.mark)
        self.axes._scenegraph.add_edge(self.axes, 'render', self.mark)

        return newmark



    # def draw_tip_box(
    #     self, 
    #     names=None, 
    #     regex=None, 
    #     wildcard=None, 
    #     yspace=None, 
    #     xspace=None, 
    #     **kwargs):
    #     """
    #     Draw a rectangle around the tips of a clade on a toytree.

    #     Parameters:
    #     -----------
    #     names, regex, wildcard:
    #         Choose one of these three methods to select one or more tipnames. 
    #         The clade composing all descendants of their common ancestor will 
    #         be highlighted.

    #     yspace (float or None):
    #         The extent to which boxes extend above and below the root and tip
    #         nodes. If None then this is automatically generated.

    #     xspace (float or None):
    #         The extent to which the clade box extends to the sides 
    #         (out of the clade towards other tips.) If None default uses 0.5.

    #     kwargs:
    #         Additional styling options are supported: color, opacity, etc.

    #     Returns:
    #     ------------
    #     Toyplot.mark.Range
    #     """

    #     # get the common ancestor
    #     nidx = self.tree.get_mrca_idx_from_tip_labels(
    #         names=names, regex=regex, wildcard=wildcard)

    #     # get tips descended from mrca
    #     tips = self.tree.idx_dict[nidx].get_leaves()
    #     tidxs = [i.idx for i in tips]

    #     # get nudge size from dists in the tree or user supplied
    #     if not yspace:
    #         yspace = self.tree.get_node_values("dist", 1, 1).mean() / 4.
    #     if not xspace:
    #         xspace = 0.5

    #     # distance in PIXELS to the tip labels
    #     tipx = toyplot.units.convert(
    #         mark.tip_labels_style["-toyplot-anchor-shift"], 'px')

    #     # left and right positions
    #     if self.mark.layout == 'r':           

    #         # get unit conversion
    #         tipstart = tipx / (axes.project('x', 1) - axes.project('x', 0))
    #         xmin = self.mark.ntable[nidx, 0] - yspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + yspace
    #         ymin = min(self.mark.ntable[tidxs, 1]) - xspace
    #         ymax = max(self.mark.ntable[tidxs, 1]) + xspace   

    #     if self.mark.layout == 'l':
    #         xmin = self.mark.ntable[nidx, 0] + yspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) - yspace
    #         ymin = max(self.mark.ntable[tidxs, 1]) + xspace
    #         ymax = min(self.mark.ntable[tidxs, 1]) - xspace   

    #     elif self.mark.layout == 'd':
    #         ymax = self.mark.ntable[nidx, 1] + yspace
    #         ymin = min(self.mark.ntable[tidxs, 1]) - yspace
    #         xmin = min(self.mark.ntable[tidxs, 0]) - xspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

    #     elif self.mark.layout == 'u':
    #         ymin = self.mark.ntable[nidx, 1] - yspace
    #         ymax = min(self.mark.ntable[tidxs, 1]) + yspace
    #         xmin = min(self.mark.ntable[tidxs, 0]) - xspace
    #         xmax = max(self.mark.ntable[tidxs, 0]) + xspace               

    #     # draw the rectangle
    #     mark = self.axes.rectangle(xmin, xmax, ymin, ymax, **kwargs)
    #     return mark



    # def generate_rectangle(self, firstname=None, lastname=None, axes=None, color="green", opacity=.25):
    #     """
    #     Returns an updated axes with a generated rectangle based on input labels provided
    #     """
    #     index_of_first = self.get_mrca_idx_from_tip_labels(names=firstname)
    #     index_of_last =  self.get_mrca_idx_from_tip_labels(names=lastname)
    #     x_vals = (x[0] for x in self.get_node_coordinates())

    #     axes.rectangle(
    #         min(self.get_tip_coordinates()[index_of_first][0],  self.get_tip_coordinates()[index_of_last][0]),
    #         max(x_vals),
    #         self.get_tip_coordinates()[index_of_first][1],
    #         self.get_tip_coordinates()[index_of_last][1],
    #         opacity=opacity,
    #         color=color,
    #     )
    #     return axes





# class TreeInference:
# - get distance matrix (from an input data set... phy, nex)
# - ----- create a class to store DNA matrix (pandas colored)
# - NJ tree infer
#   ------ uses distance matrix
# - UPGMA tree infer
#   ------ uses distance matrix


#class TreeMoves:
#     def move_spr(self):
#         """
#         Sub-tree pruning and Regrafting. 
#         Select one edge randomly from the tree and split on that edge to create
#         two subtrees. Attach one of the subtrees (e.g., the smaller one) 
#         randomly to the larger tree to create a new node.
#         ... does SPR break edges connected to root when tree is real rooted?
#         """
#         pass
#         # On rooted trees we can work with nodes easier than edges. Start by
#         # selected a node at random that is not root.
#         # nodes = [i for i in self.ttree.tree.traverse() if not i.is_root()]
#         # rnode = nodes[random.randint(0, len(nodes) - 1)]
#         # # get all edges on the tree, skip last one which is non-real root edge
#         # edges = self.ttree.tree.get_edges()[:-1]
#         # # select a random edge
#         # redge = edges[random.randint(0, len(edges))]
#         # # break into subtrees
#         # tre1 = self.tree.prune(self.tree.get_common_ancestor(redge[0]).idx)
#         # tre2 = self.tree.prune(self.tree.get_common_ancestor(redge[1]).idx)



#     def move_tbr(self):
#         pass


#     def move_nni(self):
#         pass


#     def non_parametric_rate_smoothing(self):
#         """
#         Non-parametric rate smooting.
#         A method for estimating divergence times when evolutionary rates are 
#         variable across lineages by minimizing ancestor-descendant local rate
#         changes. According to Sanderson this method is motivated by the 
#         likelihood that evolutionary rates are autocorrelated in time.

#         returns Toytree
#         """
#         # p is a fixed exponent
#         p = 2
#         W = []
#         for node in self.ttree.traverse():
#             if not node.is_leaf():
#                 children = node.children
#                 ks = []
#                 for child in children:
#                     dist = abs(node.dist - child.dist)
#                     ks.append(dist ** p)
#                 W.append(sum(ks))

#         # root rate is mean of all descendant rates -- 
#         # n is the number of edges (rates) (nnodes - 1 for root)
#         r_root = np.mean(W)
#         rootw = []
#         for child in self.ttree.tree.children:
#             rootw.append((r_rroot - child.dist) ** p)
#         w_root = sum(rootw)
#         W.append(w_root)
#         k = []
#         for 
#         k = sum(  np.exp(abs(ri - rj), p)  )
#         W = sum(k)


#     def penalized_likelihood(...):
#         pass
#

# def wfunc(ttree, p):
#     ws = []
#     for node in ttree.tree.traverse():
#         if not node.is_leaf():          
#             w = sum([(node.dist - child.dist) ** p for child in node.children])
#             ws.append(w)
#     return sum(ws)






#######################################################
# Other
#######################################################
def bpp2newick(bppnewick):
    "converts bpp newick format to normal newick. ugh."
    regex1 = re.compile(r" #[-+]?[0-9]*\.?[0-9]*[:]")
    regex2 = re.compile(r" #[-+]?[0-9]*\.?[0-9]*[;]")
    regex3 = re.compile(r": ")
    new = regex1.sub(":", bppnewick)
    new = regex2.sub(";", new)
    new = regex3.sub(":", new)
    return new.strip()



# TODO: would be useful for (eg., root) to have option to return not mrca, 
# and fuzzy match just tips, or nodes, etc...




def normalize_values(vals, nbins=10, minsize=2, maxsize=12):
    """
    Distributes values into bins spaced at reasonable sizes for plotting.
    Example, this can be used automatically scale Ne values to plot as 
    edge widths.
    """

    # make copy of original
    ovals = deepcopy(vals)

    # if 6X min value is higher than max then add this 
    # as a fake value to scale more nicely
    vals = list(vals)
    if min(vals) * 6 > max(vals):
        vals.append(min(vals) * 6)

    # sorted vals list
    svals = sorted(vals)

    # put vals into bins
    bins = np.histogram(vals, bins=nbins)[0]

    # convert binned vals to widths in 2-12
    newvals = {}
    sizes = np.linspace(minsize, maxsize, nbins)
    for idx, inbin in enumerate(bins):
        for num in range(inbin):
            newvals[svals.pop(0)] = sizes[idx]
    return np.array([newvals[i] for i in ovals])



# def fuzzy_match_tipnames(ttree, names, wildcard, regex, mono=True, retnode=True):
def fuzzy_match_tipnames(ttree, names, wildcard, regex, mrca=True, mono=True):
    """
    Used in multiple internal functions (e.g., .root()) and .drop_tips())
    to select an internal mrca node, or multiple tipnames, using fuzzy matching
    so that every name does not need to be written out by hand.

    name: verbose list
    wildcard: matching unique string
    regex: regex expression
    mrca: return mrca node of selected tipnames. 
    mono: raise error if selected tipnames are not monophyletic    
    """
    # require arguments
    if not any([names, wildcard, regex]):
        raise ToytreeError(
            "must enter an outgroup, wildcard selector, or regex pattern")

    # get list of **nodes** from {list, wildcard, or regex}
    tips = []
    if names:
        if isinstance(names, (str, int)):
            names = [names]
        notfound = [i for i in names if i not in ttree.get_tip_labels()]
        if any(notfound):
            raise ToytreeError(
                "Sample {} is not in the tree".format(notfound))
        tips = [i for i in ttree.treenode.get_leaves() if i.name in names]

    # use regex to match tipnames
    elif regex:
        tips = [
            i for i in ttree.treenode.get_leaves() if re.match(regex, i.name)
        ]               
        if not any(tips):
            raise ToytreeError("No Samples matched the regular expression")

    # use wildcard substring matching
    elif wildcard:
        tips = [i for i in ttree.treenode.get_leaves() if wildcard in i.name]
        if not any(tips):
            raise ToytreeError("No Samples matched the wildcard")

    # build list of **tipnames** from matched nodes
    if not tips:
        raise ToytreeError("no matching tipnames")       
    tipnames = [i.name for i in tips]

    # if a single tipname matched no need to check for monophyly
    if len(tips) == 1:
        if mrca:
            return tips[0]
        else:
            return tipnames

    # if multiple nodes matched, check if they're monophyletic
    mbool, mtype, mnames = (
        ttree.treenode.check_monophyly(
            tipnames, "name", ignore_missing=True)
    )

    # get mrca node
    node = ttree.treenode.get_common_ancestor(tips)

    # raise an error if required to be monophyletic but not
    if mono:
        if not mbool:
            raise ToytreeError(
                "Taxon list cannot be paraphyletic")

    # return tips or nodes
    if not mrca:
        return tipnames
    else:
        return node



#!/usr/bin/env python

"""
A class object for generating and storing Toytree plotting coordinates.
"""

import numpy as np
from .utils import ToytreeError


"""
TODO:

- it may be useful to decompose update so plotting coords change but not idxs
-
"""


class Coords:
    """
    Generates and stores plotting coordinates for nodes and edges of a tree.
    Uses the toytree _style information (e.g., layout, use_edge_lengths).

    Attributes:
    -----------
    edges: ndarray
        2-d array of pairs showing (parent, child) relationships.
        Rows are ordered in postorder traversal order for plotting.
    verts: ndarray
        2-d array with relative cartesian coordinate positions of nodes (n)
        Rows are ordered by the numbered 'idx' feature of nodes.
        ...
    """
    def __init__(self, ttree):

        # the toytree
        self.ttree = ttree

        # the vertices and edge tuples for normal layouts ('n')
        self.edges = None
        self.verts = None


    def update(self, layout=None):
        """
        Updates cartesian coordinates for drawing tree graph
        """
        # new idx_dict (can't just overwrite existing b/c nnodes may changed)
        self.ttree.idx_dict = {}

        # store tree dimensions on update
        self.ttree.nnodes = 0
        self.ttree.ntips = 0
        for node in self.ttree.treenode.traverse():
            self.ttree.nnodes += 1
            if node.is_leaf():
                self.ttree.ntips += 1

        # updates idxs and fixed_idx for any tree manipulations
        self.update_idxs()
        self.circ = Circle(self.ttree)

        # get new edges shape and fill idx_dict
        self.edges = self.get_edges()

        # get edges and verts (node locations)
        layout = self.ttree.style.layout
        if layout == 'c':
            self.verts = self.get_radial_coords()
        else:
            self.verts = self.get_linear_coords(layout=layout)



    def update_idxs(self):
        """
        set root idx highest, then all internal nodes are numbered down
        in levelorder traversal, but tips are ordered numerically from
        bottom to top in right facing ladderized tree plot order.
        """

        # internal nodes: root is highest idx
        idx = self.ttree.nnodes - 1

        for node in self.ttree.treenode.traverse("levelorder"):
            if not node.is_leaf():
                node.add_feature("idx", idx)
                self.ttree.idx_dict[idx] = node
                if not node.name:
                    node.name = str(idx)
                idx -= 1

        # external nodes: lowest numbers are for tips (0-N)
        for node in self.ttree.treenode.get_leaves():
            node.add_feature("idx", idx)
            self.ttree.idx_dict[idx] = node
            if not node.name:
                node.name = str(idx)
            idx -= 1



    # def update_fixed_order(self):
    #     "after pruning fixed order needs update to match new nnodes/ntips."
    #     # set tips order if fixing for multi-tree plotting (default None)
    #     fixed_order = self.ttree._fixed_order
    #     self.ttree_fixed_order = None
    #     self.ttree_fixed_idx = list(range(self.ttree.ntips))

    #     # check if fixed_order changed:
    #     if fixed_order:
    #         fixed_order = [
    #             i for i in fixed_order if i in self.ttree.get_tip_labels()
    #         ]
    #         self.ttree._set_fixed_order(fixed_order)
    #     # else:
    #         # self.ttree._fixed_idx = list(range(self.ttree.ntips))



    def get_edges(self):
        """
        This could change if a tree dropped nodes.
        """
        # use cache to fill edges array
        edges = np.zeros((self.ttree.nnodes - 1, 2), dtype=int)
        for idx in range(self.ttree.nnodes - 1):
            parent = self.ttree.idx_dict[idx].up
            if parent:
                edges[idx, :] = (parent.idx, idx)
        return edges



    def get_radial_coords(self, use_edge_lengths=True):
        """
        Assign .edges and .verts for node positions in a fan tree.
        The farthest tip aligns at the circumference.
        """

        verts = np.zeros((self.ttree.nnodes, 2), dtype=float)

        # shortname
        if not use_edge_lengths:
            nbits = self.ttree.treenode.get_farthest_leaf(True)[1]

        # use cache to fill edges array
        for idx in range(self.ttree.nnodes):
            node = self.ttree.idx_dict[idx]

            # leaves: x positions are evenly spaced around circumference
            if node.is_leaf() and (not node.is_root()):

                # store radians (how far around from zero to 2pi)
                node.radians = self.circ.tip_radians[idx]

                # get positions of tips using radians and radius
                if use_edge_lengths:
                    node.radius = self.circ.radius - node.height
                    node.x, node.y = self.circ.get_node_coords(node)
                else:
                    node.radius = nbits
                    node.x, node.y = self.circ.get_node_coords(node)

            # internal nodes comes after tips. Inherit position from children.
            else:

                # height is either distance or nodes from root
                if use_edge_lengths:
                    node.radius = self.circ.radius - node.height
                else:
                    node.radius = sum(1 for i in node.iter_ancestors())
                    # max([i.radius for i in node.children]) - 1

                # x position is halfway between children x-pos
                node.radians = np.mean([i.radians for i in node.children])
                node.x, node.y = self.circ.get_node_coords(node)

            # store the x,y vertex positions
            verts[node.idx] = [node.x, node.y]
        return verts

        # scale so that node idx 0 (or fixed_order x) is at (0, 0)
        # adjust = self.ttree._coords.verts[:, 1].min()
        # self.ttree._coords.verts[:, 1] -= adjust        



    def get_linear_coords(
        self,
        layout=None,
        use_edge_lengths=True,
        fixed_order=None,
        fixed_position=None,
        ):
        """
        Sets .edges, .verts for node positions.
        X and Y positions here refer to base assumption that tree is right
        facing, reorient_coordinates() will handle re-translating this.        
        """
        verts = np.zeros((self.ttree.nnodes, 2), dtype=float)

        # store verts array with x,y positions of nodes (lengths of branches)
        # we want tips to align at the right face (larger axis number)
        fartip = self.ttree.treenode.get_farthest_leaf()[0]
        hgt = sum([
            fartip.dist,
            sum(i.dist for i in fartip.iter_ancestors() if not i.is_root())
        ])

        # finding tip distance from root is a little slow, it would be faster
        # just progress forward from root, but this wouldn't easily allow for
        # spacing fixed order tips. So instead we start from tips. 
        for idx in range(self.ttree.nnodes):
            node = self.ttree.idx_dict[idx]

            # position the x-labels
            if node.is_leaf() and (not node.is_root()):
                if use_edge_lengths:
                    toroot = sum(
                        i.dist for i in node.iter_ancestors()
                        if not i.is_root()
                    )
                    node.y = hgt - (node.dist + toroot)
                else:
                    node.y = 0.

                # order of xlabels
                if fixed_order is not None:

                    # the position is explicit
                    if fixed_position is not None:
                        node.x = fixed_position[fixed_order.index(node.name)]

                    # simply use the index as position
                    else:
                        node.x = fixed_order.index(node.name)
                else:

                    # the position is exxplicit
                    if fixed_position is not None:
                        node.x = fixed_position[node.idx]

                    # simply use the index as position
                    else:
                        node.x = node.idx

            else:
                nch = node.children
                # empty trees only
                if not nch:
                    node.x = node.y = 0

                # all other trees
                else:
                    if use_edge_lengths:
                        node.y = nch[0].y + nch[0].dist
                    else:
                        node.y = max((i.y for i in nch)) + 1

                    # internal node at midpoint of children
                    node.x = sum(i.x for i in nch) / float(len(nch))

            # store the vertex
            verts[idx] = [node.x, node.y]

        # scale so that node idx 0 (or fixed_order x) is at (0, 0)
        if use_edge_lengths:
            adjust = verts[:, 1].min()
            verts[:, 1] -= adjust

        # check layout is in directionals
        if layout not in "drlu":
            raise ToytreeError("layout not recognized")

        # default: Down-facing tips align at y=0, first ladderized tip at x=0            
        if layout == 'd':
            tmp = verts

        # right-facing tips align at x=0, last ladderized tip at y=0
        if layout == 'r':
            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(verts.shape)
            tmp[:, 1] = verts[:, 0]
            tmp[:, 0] = verts[:, 1] * -1

        elif layout == 'l':
            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(verts.shape)
            tmp[:, 1] = verts[:, 0]
            tmp[:, 0] = verts[:, 1] 

        # elif layout == 'u':
        else:
            # verts swap x and ys and make xs 0 to negative
            tmp = np.zeros(verts.shape)
            tmp[:, 1] = verts[:, 1] * -1
            tmp[:, 0] = verts[:, 0] * -1

        return tmp
       



    def get_radii(self, layout):
        """
        radius of every node relative to 0,0 origin.
        """
        # custom request layout
        if layout is None:
            layout = self.ttree.layout

        # get circular layout
        if layout == 'c':
            self.radii = np.array(
                [self.ttree.idx_dict[i].radius for i in range(self.nnodes)])
        else:
            self.radii = np.repeat(0, self.ntips)



    def get_unrooted_coords(self, layout):
        """
        Project nodes onto an unrooted layout of either type A or B.
        """
        
        # get projection
        return self






class Circle:
    """
    When init from a toytree it can return coordinates based on 
    distributing tips equally in radians from 0-2pi (TODO: allow 
    setting start/end).

    The radius for farthest tip-ends is set to the tree height, 
    and origin is at 0.0.
    """
    def __init__(self, tre):

        # set radius
        self.tre = tre
        self.radius = self.tre.treenode.height 
        # get_distance(self.tre.treenode.get_farthest_leaf()[0])

        # origin
        self.o = (0, 0)
        # self.tre.style.xbaseline, self.tre.style.ybaseline)  # -self.radius, 0)

        # tips (bottom to top) are evenly spread from 0 to -2pi (counter clock)
        self.tip_radians = np.linspace(0, -np.pi * 2, self.tre.ntips + 1)[:-1]


    def get_node_coords(self, node):
        """
        get node coord. Node has radius assigned on the fly.
        """
        x = self.o[0] + node.radius * np.cos(node.radians)
        y = self.o[1] - node.radius * np.sin(node.radians)
        return x, y


    # def get_node_lines(self, node):
    #     x = self.o[0] + node.up.radius * np.cos(node.radians)
    #     y = self.o[1] - node.up.radius * np.sin(node.radians)
    #     return x, y      


    def get_tip_end_angles(self):
        """
        node radians for printing tip labels should be from the root (0,0)
        """
        return np.array([np.rad2deg(abs(i)) for i in self.tip_radians])


    def get_tip_end_coords(self):
        """
        node tip coords must calculate new radian angle relative to parent 
        node and then add offset amount to radius when calculating (x, y).
        """
        xs = [self.o[0] + self.radius * np.cos(i) for i in self.tip_radians]        
        ys = [self.o[1] - self.radius * np.sin(i) for i in self.tip_radians]
        return np.stack([xs, ys], axis=1)


    # def get_parent_coords(self, node):
    #     """
    #     get x,y coordinates of parent based on radians of children
    #     """
    #     # calculate x, y coords
    #     x = self.o[0] + node.radius * np.cos(midradian)
    #     y = self.o[1] + node.radius * np.sin(midradian)
    #     return x, y








def move_nni_unrooted():
    raise NotImplementedError("TODO")

def iter_spr_rooted(tree: ToyTree):
    """Return a generator that will visit all trees within one SPR
    move of the current tree.

    The returned set of trees does not visit the original tree.

    NOT YET WORKING/TESTED (you can help!)

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, seed=123)
    >>> spr_trees = iter_spr_unrooted(tree)
    >>> logliks = [func(stre) for stre in spr_trees]
    """
    # just iter over each possible subtree and each possible placement,
    # right? But some are redundant? Try it out...
    # raise NotImplementedError("TODO")
    raise NotImplementedError("TODO")
    edges = tree.get_edges()
    nedges = edges[edges[:, 0] != tree.treenode.idx]

    # iterate over rows of edge matrix
    for eidx in range(nedges.shape[0]):
        nnidx = nedges[eidx, 1]
        tips = tree.get_tip_labels(nnidx)
        print(nnidx, tips)

        # iterate over possible edges to place this subtree (not orig edge)
        _sub1 = tree.drop_tips(tips)
        for nidx in range(1, _sub1.nnodes):
            if nidx in [tree.treenode.idx, nnidx]:
                continue

            # toytrees
            sub0 = tree.prune(tips)
            sub1 = tree.drop_tips(tips)
            node = sub1[nidx]

            new = toytree.Node(name="spr", dist=node.dist / 2)
            node.up.children.append(new)
            node.up.children.remove(node)
            node.up = new
            new.children.append(node)
            new.children.append(sub0.treenode)
            sub0.up = new
            yield toytree.ToyTree(sub1.treenode)


def move_spr_unrooted(tree: ToyTree, seed:Optional[int]=None):
    """Return an unrooted ToyTree after one SPR move to input tree.

    Select one edge randomly from the tree and split on that edge to
    create two subtrees. Attach one of the subtrees (e.g., the
    smaller one) randomly to the larger tree to create a new node.
    """
    # ensure tree is unrooted
    if tree.is_rooted():
        tree = tree.unroot()

    # seed generator
    rng = np.random.default_rng(seed)

    # select a random edge
    nidx = rng.integers(tree.nnodes - 1)

    # get subtree below this edge
    sub0 = tree.prune(tree.get_tip_labels(nidx))

    # get subtree with sub0 removed
    sub1 = tree.drop_tips(tree.get_tip_labels(nidx))

    # randomly select an edge on sub1 to add sub0 back to
    nidx = rng.choice(range(1, sub1.nnodes))
    node = sub1[nidx]

    # randomly select a point on the branch to add new node
    if len(sub0) == 1:
        sub0 = sub0.treenode.children[0]
    else:
        sub0 = sub0.treenode
        # node.up.children.append(sub0)

    new = toytree.Node(name="spr", dist=rng.uniform(node.dist))
    node.up.children.append(new)
    node.up.children.remove(node)
    node.up = new
    new.children.append(node)
    new.children.append(sub0)
    sub0.up = new

    tree = toytree.ToyTree(sub1.treenode)
    return tree


def move_spr_rooted(
    tree: ToyTree,
    seed: Optional[int]=None,
    inplace: bool=False,
) -> ToyTree:
    """Return a rooted ToyTree one SPR move from the current tree.

    Performs a subtree pruning and regrafting move on the input tree.
    Selects one edge randomly from the tree and splits on that edge
    to create two subtrees, then attaches one of the subtrees (the
    smaller one) randomly to the larger tree to create a new node.

    This implementation is super slow, should not use prune and
    drop_tips since these require an update call.
    """
    # seed generator
    rng = np.random.default_rng(seed)

    # randomly select one edge that is not leading to the root
    edges = tree.get_edges()
    nedges = edges[edges[:, 0] != tree.treenode.idx]
    nidx = nedges[rng.integers(nedges.shape[0]), 0]
    tips = tree.get_tip_labels(nidx)

    # get subtree below selected edge
    # TODO: much faster methods would avoid prune and drop calls.
    sub0 = tree.prune(tips)

    # get subtree with sub0 clade removed
    sub1 = tree.drop_tips(tips)

    # randomly select an edge on sub1 to add sub0 back to
    # TODO, what is this selecting? why 1:?
    nidx = rng.choice(sub1.get_nodes()[1:])
    node = sub1[nidx]

    # randomly select a point on the branch to add new node
    new = toytree.Node(name="spr", dist=rng.uniform(node.dist))
    node.up.children.append(new)
    node.up.children.remove(node)
    node.up = new
    new.children.append(node)
    new.children.append(sub0.treenode)
    sub0.up = new

    tree = toytree.ToyTree(sub1.treenode)
    return tree


def _move_spr_rooted_fast(tree:toytree.ToyTree, seed:Optional[int]=None):
    """Faster implementation that avoids ToyTree coords updates.

    Returns an rooted ToyTree where one subtree pruning and
    regrafting move has been performed from the input tree.

    Select one edge randomly from the tree and split on that edge to
    create two subtrees. Attach one of the subtrees (e.g., the
    smaller one) randomly to the larger tree to create a new node.
    """
    # seed generator
    rng = np.random.default_rng(seed)

    # randomly select one edge that is not leading to the root
    edges = tree.get_edges()
    nedges = edges[edges[:, 0] != tree.treenode.idx]
    nidx = nedges[rng.integers(nedges.shape[0]), 0]
    tips = tree.get_tip_labels(nidx)

    # get subtree below selected edge
    tree1 = tree.treenode._clone()
    tree1.prune(tips)

    # get subtree with sub0 clade removed
    tree2 = tree.treenode._clone()
    tree2.prune([i for i in tree.get_tip_labels() if i not in tips])

    # randomly select a non-root node on sub1
    node = rng.choice(list(tree2.iter_descendants()))

    # randomly select a point on the branch to add new node
    new = toytree.Node(name="spr", dist=rng.uniform(node.dist))
    node.up.children.append(new)
    node.up.children.remove(node)
    node.up = new
    new.children.append(node)
    new.children.append(tree1)
    tree1.up = new

    tree = toytree.core.tree.ToyTree(tree2)
    return tree

